# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the GSPV Approach.
"""
import copy
import numpy as np
import schedula as sh
from .cmv import CMV
import co2mpas.utils as co2_utl
from co2mpas.defaults import dfl
from .core import prediction_gears_gsm

dsp = sh.BlueDispatcher(name='Gear Shifting Power Velocity Approach')
dsp.add_data('stop_velocity', dfl.values.stop_velocity)


def _gspv_interpolate_cloud(powers, velocities):
    from sklearn.isotonic import IsotonicRegression
    from scipy.interpolate import InterpolatedUnivariateSpline
    regressor = IsotonicRegression()
    regressor.fit(powers, velocities)
    x = np.linspace(min(powers), max(powers))
    y = regressor.predict(x)
    return InterpolatedUnivariateSpline(x, y, k=1, ext=3)


# noinspection PyMissingOrEmptyDocstring,PyPep8Naming
class GSPV(CMV):
    def __init__(self, *args, cloud=None, velocity_speed_ratios=None):
        super(GSPV, self).__init__(*args)
        if args and isinstance(args[0], GSPV):
            if not cloud:
                self.cloud = args[0].cloud
            if velocity_speed_ratios:
                self.convert(velocity_speed_ratios)
            else:
                velocity_speed_ratios = args[0].velocity_speed_ratios
        else:
            self.cloud = cloud or {}

        self.velocity_speed_ratios = velocity_speed_ratios or {}
        if cloud:
            self._fit_cloud()

    def __repr__(self):
        from pprint import pformat
        s = 'GSPV(cloud={}, velocity_speed_ratios={})'
        vsr = pformat(self.velocity_speed_ratios)
        s = s.format(pformat(self.cloud), vsr)
        return s.replace('inf', "float('inf')")

    # noinspection PyMethodOverriding
    def fit(self, gears, velocities, motive_powers, velocity_speed_ratios,
            stop_velocity):
        self.clear()

        self.velocity_speed_ratios = velocity_speed_ratios

        it = zip(velocities, motive_powers, co2_utl.pairwise(gears))

        for v, p, (g0, g1) in it:
            if v > stop_velocity and g0 != g1:
                x = self.get(g0, [[], [[], []]])
                if g0 < g1 and p >= 0:
                    x[1][0].append(p)
                    x[1][1].append(v)
                elif g0 > g1 and p <= 0:
                    x[0].append(v)
                else:
                    continue
                self[g0] = x

        self[0] = [[0.0], [[0.0], [stop_velocity]]]

        self[max(self)][1] = [[0, 1], [dfl.INF] * 2]

        self.cloud = {k: copy.deepcopy(v) for k, v in self.items()}

        self._fit_cloud()

        return self

    def _fit_cloud(self):
        from scipy.interpolate import InterpolatedUnivariateSpline as spl

        def _line(n, m, i):
            x = np.mean(m[i]) if m[i] else None
            k_p = n - 1
            while k_p > 0 and k_p not in self:
                k_p -= 1
            x_up = self[k_p][not i](0) if k_p >= 0 else x

            if x is None or x > x_up:
                x = x_up
            return spl([0, 1], [x] * 2, k=1)

        self.clear()
        self.update(copy.deepcopy(self.cloud))

        for k, v in sorted(self.items()):
            v[0] = _line(k, v, 0)

            if len(v[1][0]) > 2:
                v[1] = _gspv_interpolate_cloud(*v[1])
            elif v[1][1]:
                v[1] = spl([0, 1], [np.mean(v[1][1])] * 2, k=1)
            else:
                v[1] = self[k - 1][0]

    @property
    def limits(self):
        limits = {}
        X = [dfl.INF, 0]
        for v in self.cloud.values():
            X[0] = min(min(v[1][0]), X[0])
            X[1] = max(max(v[1][0]), X[1])
        X = list(np.linspace(*X))
        X = [0] + X + [X[-1] * 1.1]
        for k, func in self.items():
            limits[k] = [(f(X), X) for f, x in zip(func, X)]
        return limits

    def fig(self):
        import itertools
        import plotly.graph_objs as go
        from plotly.colors import DEFAULT_PLOTLY_COLORS
        colors = itertools.cycle(DEFAULT_PLOTLY_COLORS)
        fig = go.Figure()
        for k, v in self.limits.items():
            color = next(colors)
            for (s, l), (x, y) in zip((('down', 'dash'), ('up', None)), v):
                if x[0] < dfl.INF:
                    fig.add_trace(go.Scatter(
                        name='Gear %d:%s-shift' % (k, s), x=x, y=y,
                        line=dict(color=color, dash=l), mode='lines'
                    ))
            cy, cx = self.cloud[k][1]
            if cx[0] < dfl.INF:
                fig.add_trace(go.Scatter(
                    x=cx, y=cy, marker=dict(color=color), mode='markers',
                    showlegend=False
                ))
        fig.update_layout(
            title=self.__class__.__name__,
            xaxis_title="Velocity [km/h]",
            yaxis_title="Power [kW]"
        )
        return fig

    def _init_gear(self, times, velocities, accelerations, motive_powers,
                   engine_coolant_temperatures):
        keys = sorted(self.keys())
        matrix, c = {}, len(keys) - 1
        from co2mpas.utils import List
        if isinstance(velocities, List) or isinstance(motive_powers, List):
            for i, k in enumerate(keys):
                matrix[k] = self[k], (keys[max(0, i - 1)], keys[min(i + 1, c)])

            def _next(gear, index):
                # noinspection PyShadowingNames
                v, p = velocities[index], motive_powers[index]
                (_down, _up), (g0, g1) = matrix[gear]
                if v >= _up(p):
                    return g1
                if v < _down(p):
                    return g0
                return gear
        else:
            r = velocities.shape[0]
            for i, g in enumerate(keys):
                down, up = [func(motive_powers) for func in self[g]]
                matrix[g] = p = np.tile(g, r)
                p[velocities < down] = keys[max(0, i - 1)]
                p[velocities >= up] = keys[min(i + 1, c)]

            def _next(gear, index):
                return matrix[gear][index]
        return _next

    # noinspection PyUnresolvedReferences,PyTypeChecker
    def convert(self, velocity_speed_ratios):
        if velocity_speed_ratios != self.velocity_speed_ratios:
            # noinspection PyProtectedMember
            from .cmv import _convert_limits
            vsr, n_vsr = self.velocity_speed_ratios, velocity_speed_ratios

            limits = [dfl.INF, 0]

            for v in self.cloud.values():
                if v[1][0]:
                    limits[0] = min(min(v[1][0]), limits[0])
                    limits[1] = max(max(v[1][0]), limits[1])

            cloud = self.cloud = {}

            for p in np.linspace(*limits):
                cmv = _convert_limits(
                    {k: [f(p) for f in v] for k, v in self.items()}, vsr, n_vsr
                )

                for k, (l, u) in sorted(cmv.items()):
                    c = cloud[k] = cloud.get(k, [[], [[], []]])
                    c[0].append(l)
                    c[1][0].append(p)
                    c[1][1].append(u)

            cloud[0] = [[0.0], [[0.0], [self[0][1](0.0)]]]
            cloud[max(cloud)][1] = [[0, 1], [dfl.INF] * 2]

            self._fit_cloud()

            self.velocity_speed_ratios = n_vsr

        return self


@sh.add_function(dsp, outputs=['GSPV'])
def calibrate_gspv(
        gears, velocities, motive_powers, velocity_speed_ratios, stop_velocity):
    """
    Identifies gear shifting power velocity matrix.

    :param gears:
        Gear vector [-].
    :type gears: numpy.array

    :param velocities:
        Vehicle velocity [km/h].
    :type velocities: numpy.array

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :param velocity_speed_ratios:
        Constant velocity speed ratios of the gear box [km/(h*RPM)].
    :type velocity_speed_ratios: dict[int | float]

    :param stop_velocity:
        Maximum velocity to consider the vehicle stopped [km/h].
    :type stop_velocity: float

    :return:
        Gear shifting power velocity matrix.
    :rtype: dict
    """

    gspv = GSPV().fit(
        gears, velocities, motive_powers, velocity_speed_ratios, stop_velocity
    )

    return gspv


# predict gears with corrected matrix velocity
dsp.add_function(
    function=prediction_gears_gsm,
    inputs=[
        'correct_gear', 'gear_filter', 'GSPV', 'times', 'velocities',
        'accelerations', 'motive_powers', 'cycle_type', 'velocity_speed_ratios'
    ],
    outputs=['gears']
)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the electric motor in P1 position.
"""
import schedula as sh

dsp = sh.BlueDispatcher(
    name='Motor P1',
    description='Models the motor P1 (motor downstream the engine).'
)


@sh.add_function(dsp, outputs=['motor_p1_maximum_power'])
def identify_motor_p1_maximum_power(motor_p1_powers):
    """
    Identify the maximum power of motor P1 [kW].

    :param motor_p1_powers:
        Power at motor P1 [kW].
    :type motor_p1_powers: numpy.array

    :return:
        Maximum power of motor P1 [kW].
    :rtype: float
    """
    from .p4 import identify_motor_p4_maximum_power as func
    return func(motor_p1_powers)


@sh.add_function(dsp, outputs=['motor_p1_maximum_torque'])
def identify_motor_p1_maximum_torque(motor_p1_torques):
    """
    Identify the maximum torque of motor P1 [N*m].

    :param motor_p1_torques:
        Torque at motor P1 [N*m].
    :type motor_p1_torques: numpy.array

    :return:
        Maximum torque of motor P1 [N*m].
    :rtype: float
    """
    from .p4 import identify_motor_p4_maximum_torque as func
    return func(motor_p1_torques)


@sh.add_function(dsp, outputs=['motor_p1_maximum_power'])
def calculate_motor_p1_maximum_power(
        motor_p1_rated_speed, motor_p1_maximum_torque):
    """
    Calculate the maximum power of motor P1 [kW].

    :param motor_p1_rated_speed:
        Rated speed of motor P1 [RPM].
    :type motor_p1_rated_speed: float

    :param motor_p1_maximum_torque:
        Maximum torque of motor P1 [N*m].
    :type motor_p1_maximum_torque: float

    :return:
        Maximum power of motor P1 [kW].
    :rtype: float
    """
    from .p4 import calculate_motor_p4_maximum_power as func
    return func(motor_p1_rated_speed, motor_p1_maximum_torque)


@sh.add_function(dsp, outputs=['motor_p1_rated_speed'])
def calculate_motor_p1_rated_speed(
        motor_p1_maximum_power, motor_p1_maximum_torque):
    """
    Calculate the rated speed of motor P1 [RPM].

    :param motor_p1_maximum_power:
        Maximum power of motor P1 [kW].
    :type motor_p1_maximum_power: float

    :param motor_p1_maximum_torque:
        Maximum torque of motor P1 [N*m].
    :type motor_p1_maximum_torque: float

    :return:
        Rated speed of motor P1 [RPM].
    :rtype: float
    """
    from .p4 import calculate_motor_p4_rated_speed as func
    return func(motor_p1_maximum_power, motor_p1_maximum_torque)


@sh.add_function(dsp, outputs=['motor_p1_maximum_torque'])
def calculate_motor_p1_maximum_torque(
        motor_p1_maximum_power, motor_p1_rated_speed):
    """
    Calculate the maximum torque of motor P1 [N*m].

    :param motor_p1_maximum_power:
        Maximum power of motor P1 [kW].
    :type motor_p1_maximum_power: float

    :param motor_p1_rated_speed:
        Rated speed of motor P1 [RPM].
    :type motor_p1_rated_speed: float

    :return:
        Maximum torque of motor P1 [N*m].
    :rtype: float
    """
    from .p4 import calculate_motor_p4_maximum_torque as func
    return func(motor_p1_maximum_power, motor_p1_rated_speed)


@sh.add_function(dsp, outputs=['motor_p1_maximum_power_function'])
def define_motor_p1_maximum_power_function(
        motor_p1_maximum_power, motor_p1_rated_speed):
    """
    Define the maximum power function of motor P1.

    :param motor_p1_maximum_power:
        Maximum power of motor P1 [kW].
    :type motor_p1_maximum_power: float

    :param motor_p1_rated_speed:
        Rated speed of motor P1 [RPM].
    :type motor_p1_rated_speed: float

    :return:
        Maximum power function of motor P1.
    :rtype: function
    """
    from .p4 import define_motor_p4_maximum_power_function as func
    return func(motor_p1_maximum_power, motor_p1_rated_speed)


@sh.add_function(dsp, outputs=['motor_p1_maximum_powers'])
def calculate_motor_p1_maximum_powers(
        motor_p1_speeds, motor_p1_maximum_power_function):
    """
    Calculate the maximum power vector of motor P1 [kW].

    :param motor_p1_speeds:
        Rotating speed of motor P1 [RPM].
    :type motor_p1_speeds: numpy.array | float

    :param motor_p1_maximum_power_function:
        Maximum power function of motor P1.
    :type motor_p1_maximum_power_function: function

    :return:
        Maximum power vector of motor P1 [kW].
    :rtype: numpy.array | float
    """
    from .p4 import calculate_motor_p4_maximum_powers as func
    return func(motor_p1_speeds, motor_p1_maximum_power_function)


@sh.add_function(dsp, outputs=['motor_p1_speed_ratio'])
def identify_motor_p1_speed_ratio(engine_speeds_out, motor_p1_speeds):
    """
    Identifies motor P1 speed ratio.

    :param engine_speeds_out:
        Engine speed [RPM].
    :type engine_speeds_out: numpy.array | float

    :param motor_p1_speeds:
        Rotating speed of motor P1 [RPM].
    :type motor_p1_speeds: numpy.array | float

    :return:
        Motor P1 speed ratio [-].
    :rtype: float
    """
    from .p4 import identify_motor_p4_speed_ratio
    return identify_motor_p4_speed_ratio(engine_speeds_out, motor_p1_speeds)


dsp.add_data('motor_p1_speed_ratio', 1, sh.inf(10, 1))


@sh.add_function(dsp, outputs=['motor_p1_speeds'])
def calculate_motor_p1_speeds(engine_speeds_out, motor_p1_speed_ratio):
    """
    Calculates rotating speed of motor P1 [RPM].

    :param engine_speeds_out:
        Engine speed [RPM].
    :type engine_speeds_out: numpy.array | float

    :param motor_p1_speed_ratio:
        Ratio between motor P1 speed and engine speed [-].
    :type motor_p1_speed_ratio: float

    :return:
        Rotating speed of motor P1 [RPM].
    :rtype: numpy.array | float
    """
    return engine_speeds_out * motor_p1_speed_ratio


@sh.add_function(dsp, inputs_kwargs=True, outputs=['engine_speeds_out'])
def calculate_engine_speeds_out(motor_p1_speeds, motor_p1_speed_ratio=1):
    """
    Calculates the engine speed [RPM].

    :param motor_p1_speeds:
        Rotating speed of motor P1 [RPM].
    :type motor_p1_speeds: numpy.array | float

    :param motor_p1_speed_ratio:
        Ratio between motor P1 speed and engine speed [-].
    :type motor_p1_speed_ratio: float

    :return:
        Engine speed [RPM].
    :rtype: numpy.array | float
    """
    return motor_p1_speeds / motor_p1_speed_ratio


@sh.add_function(dsp, outputs=['motor_p1_torques'])
def calculate_motor_p1_torques(motor_p1_powers, motor_p1_speeds):
    """
    Calculates torque at motor P1 [N*m].

    :param motor_p1_powers:
        Power at motor P1 [kW].
    :type motor_p1_powers: numpy.array | float

    :param motor_p1_speeds:
        Rotating speed of motor P1 [RPM].
    :type motor_p1_speeds: numpy.array | float

    :return:
        Torque at motor P1 [N*m].
    :rtype: numpy.array | float
    """
    from ...wheels import calculate_wheel_torques
    return calculate_wheel_torques(motor_p1_powers, motor_p1_speeds)


@sh.add_function(dsp, outputs=['motor_p1_powers'])
def calculate_motor_p1_powers(motor_p1_torques, motor_p1_speeds):
    """
    Calculates power at motor P1 [kW].

    :param motor_p1_torques:
        Torque at motor P1 [N*m].
    :type motor_p1_torques: numpy.array | float

    :param motor_p1_speeds:
        Rotating speed of motor P1 [RPM].
    :type motor_p1_speeds: numpy.array | float

    :return:
        Power at motor P1 [kW].
    :rtype: numpy.array | float
    """
    from ...wheels import calculate_wheel_powers
    return calculate_wheel_powers(motor_p1_torques, motor_p1_speeds)


@sh.add_function(dsp, outputs=['motor_p1_speeds'])
def calculate_motor_p1_speeds_v1(motor_p1_powers, motor_p1_torques):
    """
    Calculates rotating speed of motor P1 [RPM].

    :param motor_p1_powers:
        Power at motor P1 [kW].
    :type motor_p1_powers: numpy.array | float

    :param motor_p1_torques:
        Torque at motor P1 [N*m].
    :type motor_p1_torques: numpy.array | float

    :return:
        Rotating speed of motor P1 [RPM].
    :rtype: numpy.array | float
    """
    from ...wheels import calculate_wheel_torques
    return calculate_wheel_torques(motor_p1_powers, motor_p1_torques)


dsp.add_data('motor_p1_efficiency', 0.9)


@sh.add_function(dsp, outputs=['motor_p1_electric_powers'])
def calculate_motor_p1_electric_powers(motor_p1_powers, motor_p1_efficiency):
    """
    Calculates motor P1 electric power [kW].

    :param motor_p1_powers:
        Power at motor P1 [kW].
    :type motor_p1_powers: numpy.array | float

    :param motor_p1_efficiency:
        Motor P1 efficiency [-].
    :type motor_p1_efficiency: float

    :return:
        Electric power of motor P1 [kW].
    :rtype: numpy.array | float
    """
    from .p4 import calculate_motor_p4_electric_powers as func
    return func(motor_p1_powers, motor_p1_efficiency)


@sh.add_function(dsp, outputs=['motor_p1_powers'])
def calculate_motor_p1_powers_v1(motor_p1_electric_powers, motor_p1_efficiency):
    """
    Calculate motor P1 power from electric power and electric power losses [kW].

    :param motor_p1_electric_powers:
        Electric power of motor P1 [kW].
    :type motor_p1_electric_powers: numpy.array | float

    :param motor_p1_efficiency:
        Motor P1 efficiency [-].
    :type motor_p1_efficiency: float

    :return:
        Power at motor P1 [kW].
    :rtype: numpy.array | float
    """
    from .p4 import calculate_motor_p4_powers_v1 as func
    return func(motor_p1_electric_powers, motor_p1_efficiency)


dsp.add_data('has_motor_p1', False, sh.inf(10, 3))


@sh.add_function(dsp, outputs=['has_motor_p1'])
def identify_has_motor_p1(motor_p1_maximum_power):
    """
    Identify if the vehicle has a motor P1 [kW].

    :param motor_p1_maximum_power:
        Maximum power of motor P1 [kW].
    :type motor_p1_maximum_power: float

    :return:
        Has the vehicle a motor in P1?
    :rtype: bool
    """
    from .p4 import identify_has_motor_p4 as func
    return func(motor_p1_maximum_power)


@sh.add_function(dsp, outputs=['motor_p1_powers'])
def default_motor_p1_powers(times, has_motor_p1):
    """
    Return zero power if the vehicle has not a motor P1 [kW].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param has_motor_p1:
        Has the vehicle a motor in P1?
    :type has_motor_p1: bool

    :return:
        Power at motor P1 [kW].
    :rtype: numpy.array
    """
    from .p4 import default_motor_p4_powers as func
    return func(times, has_motor_p1)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to bypass the gear box.
"""
import schedula as sh
from co2mpas.defaults import dfl
from .cvt import (
    default_correct_gear, identify_max_speed_velocity_ratio, predict_gears, CVT
)

dsp = sh.BlueDispatcher(
    name='empty model', description='Bypass for the gear box.'
)

dsp.add_func(default_correct_gear, outputs=['correct_gear'])
dsp.add_data('max_gear', 1)
dsp.add_func(predict_gears, outputs=['gears'])
dsp.add_data('stop_velocity', dfl.values.stop_velocity)
dsp.add_func(
    identify_max_speed_velocity_ratio, outputs=['max_speed_velocity_ratio']
)


@sh.add_function(dsp, outputs=['gear_shifting_model'])
def define_gear_shifting_model():
    """
    Return a fake gear shifting model.

    :return:
        Gear shifting model.
    :rtype: CVT
    """
    return CVT()


dsp.add_function(
    function_id='calculate_gear_box_speeds_in',
    function=sh.bypass,
    inputs=['gear_box_speeds_out'],
    outputs=['gear_box_speeds_in']
)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions to convert CO2MPAS output report to DataFrames.
"""
import regex
import functools
import itertools
import logging
import collections
import schedula as sh

log = logging.getLogger(__name__)
_re_units = regex.compile(r'(\[.*\])')


def _parse_name(name, _standard_names=None):
    """
    Parses a column/row name.

    :param name:
        Name to be parsed.
    :type name: str

    :return:
        The parsed name.
    :rtype: str
    """

    if _standard_names and name in _standard_names:
        return _standard_names[name]

    return name.replace('_', ' ').capitalize()


@functools.lru_cache(None)
def _get_doc_description():
    from ..model.physical import dsp

    doc_descriptions = {}

    d = dsp.register(memo={})
    for k, v in d.data_nodes.items():
        if k in doc_descriptions or v['type'] != 'data':
            continue
        des = d.get_node(k, node_attr='description')[0]
        if not des or len(des.split(' ')) > 4:

            unit = _re_units.search(des)
            if unit:
                unit = ' %s' % unit.group()
            else:
                unit = ''
            doc_descriptions[k] = '%s%s.' % (_parse_name(k), unit)
        else:
            doc_descriptions[k] = des
    return doc_descriptions


def _param_parts(param_id):
    # noinspection PyProtectedMember
    from ..load.excel import _re_params_name
    match = _re_params_name.match(param_id).groupdict().items()
    return {i: regex.sub(r"[\W]", "_", (j or '').lower()) for i, j in match}


def _time_series2df(data, data_descriptions):
    df = collections.OrderedDict()
    for k, v in data.items():
        df[(_parse_name(_param_parts(k)['param'], data_descriptions), k)] = v
    import pandas as pd
    return pd.DataFrame(df)


def _parameters2df(data, data_descriptions, write_validator):
    import schema
    df = []
    for k, v in data.items():
        try:
            param_id, vl = write_validator(_param_parts(k)['param'], v)
            if vl is not sh.NONE:
                df.append({
                    'Parameter': _parse_name(param_id, data_descriptions),
                    'Model Name': k,
                    'Value': vl
                })
        except schema.SchemaError as ex:
            raise ValueError(k, v, ex)

    if df:
        import pandas as pd
        df = pd.DataFrame(df)
        df.set_index(['Parameter', 'Model Name'], inplace=True)
        return df
    else:
        return None


def _cycle2df(data):
    res, out = {}, data.get('output', {})
    from .excel import _sheet_name
    from ..load.schema import define_data_validation
    write_validator = define_data_validation(read=False)
    data_descriptions = _get_doc_description()
    for k, v in sh.stack_nested_keys(out, key=('output',), depth=3):
        n, k = _sheet_name(k), k[-1]
        if 'ts' == k:
            df = _time_series2df(v, data_descriptions)
        elif 'pa' == k:
            df = _parameters2df(v, data_descriptions, write_validator)
        else:
            continue

        if df is not None:
            res[n] = df
    return res


def _dd2df(dd, index=None, depth=0, col_key=None, row_key=None):
    frames = []
    import inspect
    import pandas as pd
    for k, v in sh.stack_nested_keys(dd, depth=depth):
        df = pd.DataFrame(v)
        df.drop_duplicates(subset=index, inplace=True)
        if index is not None:
            df.set_index(index, inplace=True)

        df.columns = pd.MultiIndex.from_tuples([k + (i,) for i in df.columns])
        frames.append(df)

    df = pd.concat(
        frames, copy=False, axis=1, sort=False, verify_integrity=True
    )

    if col_key is not None:
        ax = sorted(df.columns, key=col_key)
        if isinstance(df.columns, pd.MultiIndex):
            ax = pd.MultiIndex.from_tuples(ax)

        # noinspection PyUnresolvedReferences
        df = df.reindex(ax, axis='columns', copy=False)

    if row_key is not None:
        ax = sorted(df.index, key=row_key)
        if isinstance(df.index, pd.MultiIndex):
            ax = pd.MultiIndex.from_tuples(ax)
        df = df.reindex(ax, axis='index', copy=False)

    if index is not None:
        if len(index) == 1 and isinstance(df.index, pd.MultiIndex):
            df.index = pd.Index(df.index, name=index[0])
        else:
            df.index.set_names(index, inplace=True)

    return df


@functools.lru_cache(None)
def _param_orders():
    x = ('declared_co2_emission', 'co2_emission', 'fuel_consumption')
    y = ('low', 'medium', 'high', 'extra_high', 'UDC', 'EUDC', 'value')
    param = x + tuple(map('_'.join, itertools.product(x, y))) + ('status',)

    param += (
        'av_velocities', 'distance', 'init_temp', 'av_temp', 'end_temp',
        'av_vel_pos_mov_pow', 'av_pos_motive_powers',
        'av_missing_powers_pos_pow', 'sec_pos_mov_pow', 'av_neg_motive_powers',
        'sec_neg_mov_pow', 'av_pos_accelerations',
        'av_engine_speeds_out_pos_pow', 'av_pos_engine_powers_out',
        'engine_bmep_pos_pow', 'mean_piston_speed_pos_pow', 'fuel_mep_pos_pow',
        'fuel_consumption_pos_pow', 'willans_a', 'willans_b',
        'specific_fuel_consumption', 'indicated_efficiency',
        'willans_efficiency', 'times'
    )

    _map = {
        'scope': ('plan', 'flag', 'base'),
        'usage': ('target', 'output', 'input', 'data'),
        'stage': ('precondition', 'prediction', 'calibration'),
        'cycle': (
            'all', 'nedc_h', 'nedc_l', 'wltp_h', 'wltp_l','wltp_m', 'wltp_p'
        ),
        'type': ('pa', 'ts', 'pl', 'mt'),
        'param': param
    }
    _map = {k: {j: str(i).zfill(3) for i, j in enumerate(v)}
            for k, v in _map.items()}

    return _map


# noinspection PyShadowingBuiltins
def _match_part(map, *parts, default=None):
    part = parts[-1]
    try:
        return map[part],
    except KeyError:
        for k, v in sorted(map.items(), key=lambda x: x[1]):
            if k in part:
                return v, 0, part
        part = part if default is None else default
        if len(parts) <= 1:
            return max(map.values()) if map else None, 1, part
        return _match_part(map, *parts[:-1], default=part)


def _sort_key(
        parts, score_map=None,
        p_keys=('scope', 'param', 'cycle', 'usage', 'stage', 'type')):
    score_map = score_map or _param_orders()
    it = itertools.zip_longest(parts, p_keys, fillvalue=None)
    return tuple(_match_part(score_map.get(k, {}), p) for p, k in it)


def _scores2df(data):
    n = ('data', 'calibration', 'model_scores')
    if not sh.are_in_nested_dicts(data, *n):
        return {}

    scores = sh.get_nested_dicts(data, *n)

    it = (('model_selections', ['model_id'], 2, ('stage', 'cycle'), ()),
          ('score_by_model', ['model_id'], 1, ('cycle',), ()),
          ('scores', ['model_id', 'param_id'], 2, ('cycle', 'cycle'), ()),
          ('param_selections', ['param_id'], 2, ('stage', 'cycle'), ()))
    dfs = []
    for k, idx, depth, col_keys, row_keys in it:
        if k not in scores:
            continue
        df = _dd2df(
            scores[k], idx, depth=depth,
            col_key=functools.partial(_sort_key, p_keys=col_keys),
            row_key=functools.partial(_sort_key, p_keys=row_keys)
        )
        setattr(df, 'name', k)
        dfs.append(df)
    if dfs:
        return {'.'.join(n): dfs}
    else:
        return {}


@functools.lru_cache(None)
def _summary_map(short=True):
    keys = (
        'declared_sustaining_co2_emission_value', 'declared_co2_emission_value',
        'corrected_sustaining_co2_emission_value', 'co2_emission_value',
        'corrected_co2_emission_value',
    )
    if short:
        _map = {k: 'value' for k in keys}
    else:
        _map = {k: k.replace('co2_emission_', '') for k in keys}

    _map.update({
        'co2_params a': 'a',
        'co2_params a2': 'a2',
        'co2_params b': 'b',
        'co2_params b2': 'b2',
        'co2_params c': 'c',
        'co2_params l': 'l',
        'co2_params l2': 'l2',
        'co2_params t0': 't0',
        'co2_params dt': 'dt',
        'co2_params t1': 't1',
        'co2_params trg': 'trg',
        'co2_emission_low': 'low',
        'co2_emission_medium': 'medium',
        'co2_emission_high': 'high',
        'co2_emission_extra_high': 'extra_high',
        'co2_emission_UDC': 'UDC',
        'co2_emission_EUDC': 'EUDC',
        'vehicle_mass': 'mass',
    })

    return _map


def _search_unit(units, default, *keys):
    try:
        return units[keys[-1]]
    except KeyError:
        try:
            return _search_unit(units, sh.EMPTY, *keys[:-1])
        except IndexError:
            if default is sh.EMPTY:
                raise IndexError
            for i, u in units.items():
                if any(i in k for k in keys):
                    return u
            return default


@functools.lru_cache(None)
def _param_units():
    units = (
        (k, _re_units.search(v)) for k, v in _get_doc_description().items()
    )
    units = {k: v.group() for k, v in units if v}
    units.update({
        'co2_params a': '[-]',
        'co2_params b': '[s/m]',
        'co2_params c': '[(s/m)^2]',
        'co2_params a2': '[1/bar]',
        'co2_params b2': '[s/(bar*m)]',
        'co2_params l': '[bar]',
        'co2_params l2': '[bar*(s/m)^2]',
        'co2_params t': '[-]',
        'co2_params trg': '[°C]',
        'fuel_consumption': '[l/100km]',
        'co2_emission': '[CO2g/km]',
        'declared_co2_emission_value': '[CO2g/km]',
        'av_velocities': '[kw/h]',
        'av_vel_pos_mov_pow': '[kw/h]',
        'av_pos_motive_powers': '[kW]',
        'av_neg_motive_powers': '[kW]',
        'distance': '[km]',
        'init_temp': '[°C]',
        'av_temp': '[°C]',
        'end_temp': '[°C]',
        'sec_pos_mov_pow': '[s]',
        'sec_neg_mov_pow': '[s]',
        'av_pos_accelerations': '[m/s2]',
        'av_engine_speeds_out_pos_pow': '[RPM]',
        'av_pos_engine_powers_out': '[kW]',
        'engine_bmep_pos_pow': '[bar]',
        'mean_piston_speed_pos_pow': '[m/s]',
        'fuel_mep_pos_pow': '[bar]',
        'fuel_consumption_pos_pow': '[g/sec]',
        'willans_a': '[g/kW]',
        'willans_b': '[g]',
        'specific_fuel_consumption': '[g/kWh]',
        'indicated_efficiency': '[-]',
        'willans_efficiency': '[-]',
    })

    return units


def _add_units(gen, default=' ', short=True):
    p_map = _summary_map(short=short).get
    units = functools.partial(_search_unit, _param_units(), default)
    return [k[:-1] + (p_map(k[-1], k[-1]), units(*k)) for k in gen]


def _summary2df(data):
    res = []
    summary = data.get('summary', {})

    if 'results' in summary:
        r = {}
        index = ['cycle', 'stage', 'usage']

        for k, v in sh.stack_nested_keys(summary['results'], depth=4):
            l = sh.get_nested_dicts(r, k[0], default=list)
            l.append(sh.combine_dicts(sh.map_list(index, *k[1:]), v))

        if r:
            df = _dd2df(
                r, index=index, depth=2,
                col_key=functools.partial(_sort_key, p_keys=('param',) * 2),
                row_key=functools.partial(_sort_key, p_keys=index)
            )
            import pandas as pd
            df.columns = pd.MultiIndex.from_tuples(_add_units(df.columns))
            setattr(df, 'name', 'results')
            res.append(df)

    if 'selection' in summary:
        df = _dd2df(
            summary['selection'], ['model_id'], depth=2,
            col_key=functools.partial(_sort_key, p_keys=('stage', 'cycle')),
            row_key=functools.partial(_sort_key, p_keys=())
        )
        setattr(df, 'name', 'selection')
        res.append(df)

    if 'comparison' in summary:
        r = {}
        for k, v in sh.stack_nested_keys(summary['comparison'], depth=3):
            v = sh.combine_dicts(v, base={'param_id': k[-1]})
            sh.get_nested_dicts(r, *k[:-1], default=list).append(v)
        if r:
            df = _dd2df(
                r, ['param_id'], depth=2,
                col_key=functools.partial(_sort_key, p_keys=('stage', 'cycle')),
                row_key=functools.partial(_sort_key, p_keys=())
            )
            setattr(df, 'name', 'comparison')
            res.append(df)

    if res:
        return {'summary': res}
    return {}


def _co2mpas_info2df(start_time, main_flags=None):
    import socket
    import datetime
    from co2mpas import __version__
    from ..load.schema import define_flags_validation
    time_elapsed = (datetime.datetime.today() - start_time).total_seconds()
    hostname = socket.gethostname()
    info = [
        ('CO2MPAS version', __version__),
        ('Simulation started', start_time.strftime('%Y/%m/%d-%H:%M:%S')),
        ('Time elapsed', '%.3f sec' % time_elapsed),
        ('Hostname', hostname),
    ]

    if main_flags:
        validate = define_flags_validation(read=False)
        main_flags = dict(validate(k, v) for k, v in main_flags.items())
        info.extend(sorted(main_flags.items()))
    import pandas as pd
    df = pd.DataFrame(info, columns=['Parameter', 'Value'])
    df.set_index(['Parameter'], inplace=True)
    setattr(df, 'name', 'info')
    return df


_re_list = regex.compile(br'^[^\[]*(?P<list>\[.*\])[^\]]*$', regex.DOTALL)


@functools.lru_cache()
def _get_installed_packages():
    import json
    from subprocess import check_output
    try:
        out = check_output("conda list --json".split())
        m = _re_list.match(out)
        if m:
            return json.loads(m.group('list'))
        msg = 'Invalid JSON!'
    except Exception as ex:
        msg = ex
    log.info("Failed collecting installation info.\n%s", msg)


def _pipe2list(pipe, i=0, source=()):
    res, max_l = [], i
    idx = {'nodes L%d' % i: str(v) for i, v in enumerate(source)}
    node_id = 'nodes L%d' % i
    for k, v in pipe.items():
        k = sh.stlp(k)
        d = {node_id: str(k)}

        if 'error' in v:
            d['error'] = v['error']

        j, s = v['task'][2]
        n = s.workflow.nodes.get(j, {})
        if 'duration' in n:
            d['duration'] = n['duration']

        d.update(idx)
        res.append(d)

        if 'sub_pipe' in v:
            l, ml = _pipe2list(v['sub_pipe'], i=i + 1, source=source + (k,))
            max_l = max(max_l, ml)
            res.extend(l)

    return res, max_l


def _proc_info2df(data, start_time, main_flags):
    res = _co2mpas_info2df(start_time, main_flags),

    df = _get_installed_packages()
    if df:
        import pandas as pd
        df = pd.DataFrame(df).set_index('name')
        setattr(df, 'name', 'packages')
        res += df,

    df, max_l = _pipe2list(data.get('pipe', {}))

    if df:
        import pandas as pd
        df = pd.DataFrame(df)
        setattr(df, 'name', 'pipe')
        res += df,

    return {'proc_info': res}


def _dice2df(dice):
    if dice:
        import pandas as pd
        df = pd.DataFrame(sorted(dice.items()), columns=['Parameter', 'Value'])
        df.set_index(['Parameter'], inplace=True)
        setattr(df, 'name', 'dice')
        return {'dice': [df]}
    return {}


def convert2df(report, start_time, flag, dice):
    """
    Convert vehicle output report to DataFrames.

    :param report:
        Vehicle output report.
    :type report: dict

    :param start_time:
        Run start time.
    :type start_time: datetime.datetime

    :param flag:
        Command line flags.
    :type flag: dict

    :param dice:
        DICE data.
    :type dice: dict

    :return:
        DataFrames of vehicle output report.
    :rtype: dict[str, pandas.DataFrame]
    """
    res = {'graphs.%s' % k: v for k, v in report.get('graphs', {}).items()}

    res.update(_cycle2df(report))

    res.update(_scores2df(report))

    res.update(_summary2df(report))

    res.update(_dice2df(dice))

    res.update(_proc_info2df(report, start_time, flag or {}))

    res['summary'] = [res['proc_info'][0]] + res.get('summary', [])

    return res



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the DT Approach.
"""
import numpy as np
import schedula as sh
from .cmv import CMV
from .core import prediction_gears_gsm as _prediction_gears_gsm

dsp = sh.BlueDispatcher(name='Decision Tree Approach')


# noinspection PyMissingOrEmptyDocstring,PyCallByClass,PyUnusedLocal
# noinspection PyTypeChecker,PyPep8Naming
class DTGS:
    def __init__(self, velocity_speed_ratios):
        from xgboost import XGBClassifier
        self.tree = XGBClassifier()
        self.model = self.gears = None
        self.velocity_speed_ratios = velocity_speed_ratios

    def fit(self, gears, velocities, accelerations, motive_powers,
            engine_temperatures):
        i = np.arange(-1, gears.shape[0] - 1)
        i[0] = 0
        # noinspection PyProtectedMember
        from ...engine._thermal import _SelectFromModel
        from sklearn.pipeline import Pipeline
        model = self.tree
        self.model = Pipeline([
            ('feature_selection', _SelectFromModel(
                model, threshold='0.8*median', in_mask=(0, 1)
            )),
            ('classification', model)
        ])
        X = np.column_stack((
            gears[i], velocities, accelerations, motive_powers,
            engine_temperatures
        ))
        self.model.fit(X, gears)

        self.gears = np.unique(gears)
        return self

    def _init_gear(self, times, velocities, accelerations, motive_powers,
                   engine_temperatures):
        from co2mpas.utils import List
        predict = self.model.predict
        pars = (
            velocities, accelerations, motive_powers,
            engine_temperatures
        )

        if any(isinstance(v, List) for v in pars) or np.isnan(pars[-1]).any():
            x = np.empty((1, 5), float)

            def _next(gear, i):
                x[:, :] = (
                    gear, velocities[i], accelerations[i], motive_powers[i],
                    engine_temperatures[i]
                )
                return predict(x)[0]
        else:
            matrix = {}
            x = np.column_stack((np.empty_like(velocities),) + pars)
            for g in self.velocity_speed_ratios.keys():
                x[:, 0] = g
                matrix[g] = predict(x)
            del x

            def _next(gear, index):
                return matrix[gear][index]
        return _next

    def init_gear(self, *args, **kwargs):
        return CMV.init_gear(self, *args, **kwargs)

    def init_speed(self, *args, **kwargs):
        return CMV.init_speed(self, *args, **kwargs)

    def predict(self, *args, **kwargs):
        return CMV.predict(self, *args, **kwargs)

    def convert(self, velocity_speed_ratios):
        self.velocity_speed_ratios = velocity_speed_ratios
        return self


@sh.add_function(dsp, outputs=['DTGS'])
def calibrate_gear_shifting_decision_tree(
        velocity_speed_ratios, gears, velocities, accelerations, motive_powers,
        engine_temperatures):
    """
    Calibrates a decision tree to predict gears.

    :param velocity_speed_ratios:
        Constant velocity speed ratios of the gear box [km/(h*RPM)].
    :type velocity_speed_ratios: dict[int | float]

    :param gears:
        Gear vector [-].
    :type gears: numpy.array

    :param velocities:
        Vehicle velocity [km/h].
    :type velocities: numpy.array

    :param accelerations:
        Vehicle acceleration [m/s2].
    :type accelerations: numpy.array

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :param engine_temperatures:
        Engine coolant temperature vector [°C].
    :type engine_temperatures: numpy.array

    :returns:
        A decision tree to predict gears.
    :rtype: DTGS
    """

    model = DTGS(velocity_speed_ratios).fit(
        gears, velocities, accelerations, motive_powers,
        engine_temperatures
    )
    return model


def prediction_gears_gsm(*a):
    """
    Predicts gears with a gear shifting model (cmv or gspv or dtgs or mgs) [-].

    :param a:
        Arguments of `_prediction_gears_gsm`.
    :type a: tuple

    :return:
        Predicted gears.
    :rtype: numpy.array
    """
    return _prediction_gears_gsm(*a[:-1], engine_temperatures=a[-1])


prediction_gears_gsm.__doc__ = _prediction_gears_gsm.__doc__
dsp.add_function(
    function=prediction_gears_gsm,
    inputs=[
        'correct_gear', 'gear_filter', 'DTGS', 'times', 'velocities',
        'accelerations', 'motive_powers', 'engine_temperatures'
    ],
    outputs=['gears']
)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the Energy Management System.
"""
import numpy as np
import schedula as sh
from co2mpas.defaults import dfl
import co2mpas.utils as co2_utl
from numbers import Number

dsp = sh.BlueDispatcher(
    name='ems',
    description='Models the Energy Management System.'
)

dsp.add_function(
    function_id='define_motors_efficiencies',
    function=sh.bypass,
    inputs=[
        'motor_p4_front_efficiency', 'motor_p4_rear_efficiency',
        'motor_p3_front_efficiency', 'motor_p3_rear_efficiency',
        'motor_p2_efficiency', 'motor_p2_planetary_efficiency',
        'motor_p1_efficiency', 'motor_p0_efficiency'
    ],
    outputs=['motors_efficiencies']
)

dsp.add_function(
    function_id='define_motors_maximum_powers',
    function=sh.bypass,
    inputs=[
        'motor_p4_front_maximum_power', 'motor_p4_rear_maximum_power',
        'motor_p3_front_maximum_power', 'motor_p3_rear_maximum_power',
        'motor_p2_maximum_power', 'motor_p2_planetary_maximum_power',
        'motor_p1_maximum_power', 'motor_p0_maximum_power'
    ],
    outputs=['motors_maximum_powers']
)


@sh.add_function(dsp, outputs=['motors_maximums_powers'])
def define_motors_maximums_powers(
        motor_p4_front_maximum_powers, motor_p4_rear_maximum_powers,
        motor_p3_front_maximum_powers, motor_p3_rear_maximum_powers,
        motor_p2_maximum_powers, engine_speeds_parallel, final_drive_speeds_in,
        motor_p2_planetary_maximum_power_function, planetary_ratio,
        motor_p2_planetary_speed_ratio, motor_p1_maximum_power_function,
        motor_p1_speed_ratio, motor_p0_maximum_power_function,
        motor_p0_speed_ratio):
    """
    Defines maximum powers of electric motors [kW].

    :param motor_p4_front_maximum_powers:
        Maximum power vector of motor P4 front [kW].
    :type motor_p4_front_maximum_powers: numpy.array

    :param motor_p4_rear_maximum_powers:
        Maximum power vector of motor P4 rear [kW].
    :type motor_p4_rear_maximum_powers: numpy.array

    :param motor_p3_front_maximum_powers:
        Maximum power vector of motor P3 front [kW].
    :type motor_p3_front_maximum_powers: numpy.array

    :param motor_p3_rear_maximum_powers:
        Maximum power vector of motor P3 rear [kW].
    :type motor_p3_rear_maximum_powers: numpy.array

    :param motor_p2_maximum_powers:
        Maximum power vector of motor P2 [kW].
    :type motor_p2_maximum_powers: numpy.array

    :param engine_speeds_parallel:
        Hypothetical engine speed in parallel mode [RPM].
    :type engine_speeds_parallel: numpy.array

    :param final_drive_speeds_in:
        Final drive speed in [RPM].
    :type final_drive_speeds_in: numpy.array

    :param motor_p2_planetary_maximum_power_function:
        Maximum power function of planetary motor P2.
    :type motor_p2_planetary_maximum_power_function: function

    :param planetary_ratio:
        Fundamental planetary speed ratio [-].
    :type planetary_ratio: float

    :param motor_p2_planetary_speed_ratio:
        Ratio between planetary motor P2 speed and planetary speed [-].
    :type motor_p2_planetary_speed_ratio: float

    :param motor_p1_maximum_power_function:
        Maximum power function of motor P1.
    :type motor_p1_maximum_power_function: function

    :param motor_p1_speed_ratio:
        Ratio between motor P1 speed and engine speed [-].
    :type motor_p1_speed_ratio: float

    :param motor_p0_maximum_power_function:
        Maximum power function of motor P0.
    :type motor_p0_maximum_power_function: function

    :param motor_p0_speed_ratio:
        Ratio between motor P0 speed and engine speed [-].
    :type motor_p0_speed_ratio: float

    :return:
        Maximum powers of electric motors [kW].
    :rtype: numpy.array
    """
    es, p2_powers = engine_speeds_parallel, motor_p2_maximum_powers
    planet_f = motor_p2_planetary_maximum_power_function
    from ..electrics.motors.planet import calculate_planetary_speeds_in as func
    ps = func(es, final_drive_speeds_in, planetary_ratio)
    return np.column_stack((
        motor_p4_front_maximum_powers, motor_p4_rear_maximum_powers,
        motor_p3_front_maximum_powers, motor_p3_rear_maximum_powers, p2_powers,
        planet_f(ps * motor_p2_planetary_speed_ratio),
        motor_p1_maximum_power_function(es * motor_p1_speed_ratio),
        motor_p0_maximum_power_function(es * motor_p0_speed_ratio)
    ))


@sh.add_function(dsp, outputs=['drive_line_efficiencies'])
def define_drive_line_efficiencies(
        final_drive_mean_efficiency, gear_box_mean_efficiency,
        clutch_tc_mean_efficiency, planetary_mean_efficiency,
        belt_mean_efficiency):
    """
    Defines drive line efficiencies vector.

    :param final_drive_mean_efficiency:
        Final drive mean efficiency [-].
    :type final_drive_mean_efficiency: float

    :param gear_box_mean_efficiency:
        Gear box mean efficiency [-].
    :type gear_box_mean_efficiency: float

    :param clutch_tc_mean_efficiency:
        Clutch or torque converter mean efficiency [-].
    :type clutch_tc_mean_efficiency: float

    :param planetary_mean_efficiency:
        Planetary mean efficiency [-].
    :type planetary_mean_efficiency: float

    :param belt_mean_efficiency:
        Belt mean efficiency [-].
    :type belt_mean_efficiency: float

    :return:
        Drive line efficiencies vector.
    :rtype: tuple[float]
    """
    return (
        1.0, final_drive_mean_efficiency, gear_box_mean_efficiency,
        planetary_mean_efficiency, clutch_tc_mean_efficiency,
        belt_mean_efficiency
    )


@sh.add_function(dsp, outputs=['engine_speeds_parallel'], weight=sh.inf(11, 0))
def calculate_engine_speeds_parallel(gear_box_speeds_in, idle_engine_speed):
    """
    Calculate hypothetical engine speed in parallel mode [RPM].

    :param gear_box_speeds_in:
        Gear box speed [RPM].
    :type gear_box_speeds_in: numpy.array | float

    :param idle_engine_speed:
        Idle engine speed and its standard deviation [RPM].
    :type idle_engine_speed: (float, float)

    :return:
        Hypothetical engine speed in parallel mode [RPM].
    :rtype: numpy.array | float
    """
    return np.maximum(gear_box_speeds_in, idle_engine_speed[0])


@sh.add_function(dsp, outputs=['engine_speeds_parallel'])
def calculate_engine_speeds_parallel_v1(
        on_engine, engine_speeds_out, gear_box_speeds_in, idle_engine_speed):
    """
    Calculate hypothetical engine speed [RPM].

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param engine_speeds_out:
        Engine speed [RPM].
    :type engine_speeds_out: numpy.array

    :param gear_box_speeds_in:
        Gear box speed [RPM].
    :type gear_box_speeds_in: numpy.array | float

    :param idle_engine_speed:
        Idle engine speed and its standard deviation [RPM].
    :type idle_engine_speed: (float, float)

    :return:
        Hypothetical engine speed in parallel mode [RPM].
    :rtype: numpy.array | float
    """
    s = calculate_engine_speeds_parallel(gear_box_speeds_in, idle_engine_speed)
    return np.where(on_engine, engine_speeds_out, s)


def _invert(y, xp, fp):
    with np.errstate(divide='ignore', invalid='ignore'):
        x = xp[:-1] + (np.diff(xp) / np.diff(fp)) * (y[:, None] - fp[:-1])
        b = (xp[:-1] - dfl.EPS <= x) & (x <= (xp[1:] + dfl.EPS))
        # noinspection PyUnresolvedReferences
        return x[np.arange(b.shape[0]), np.nanargmax(b, 1)], y


# noinspection PyMissingOrEmptyDocstring
class FuelMapModel:
    def __init__(self, fuel_map, full_load_curve):
        from scipy.interpolate import UnivariateSpline as Spl
        from scipy.interpolate import RegularGridInterpolator as Interpolator
        self.full_load_curve = full_load_curve
        self.fc = Interpolator(
            (fuel_map['speed'], fuel_map['power']), fuel_map['fuel'],
            bounds_error=False, fill_value=0
        )
        (s, p), fc = self.fc.grid, self.fc.values

        with np.errstate(divide='ignore', invalid='ignore'):
            e = np.maximum(0, p / fc)
        e[(p > full_load_curve(s)[:, None]) | (p < 0)] = np.nan
        b = ~np.isnan(e).all(1)
        (s, i), e = np.unique(s[b], return_index=True), e[b]
        b = ~np.isnan(e).all(0)
        p = p[b][np.nanargmax(e[:, b], 1)][i]

        func = Spl(s, p, w=1 / np.clip(p * .01, dfl.EPS, 1))
        s = np.unique(np.append(s, np.linspace(s.min(), s.max(), 1000)))
        p = func(s)
        self.max_power = p.max()
        self.speed_power = Spl(s, p, s=0)
        self.power_speed = Spl(*_invert(np.unique(p), s, p)[::-1], s=0, ext=3)
        self.idle_fc = self.fc((self.power_speed(0), 0))

    def __call__(self, speed, power):
        return self.fc((speed, power))

    def fig(self):
        import plotly.graph_objs as go
        (s, p), fc = self.fc.grid, self.fc.values
        x, y = self.power_speed(p), self.full_load_curve(s)
        xp = self.full_load_curve.keywords['xp']
        fp = self.full_load_curve.keywords['fp']
        fig = go.Figure(data=[
            go.Scatter3d(
                z=np.nan_to_num(self(s, y)), x=s, y=y, mode='lines',
                line=dict(color='red', width=10)
            ),
            go.Scatter3d(
                z=np.nan_to_num(self(xp, fp)), x=xp, y=fp, mode='markers',
                marker=dict(color='red', size=10)
            ),
            go.Scatter3d(
                z=np.nan_to_num(self(x, p)), x=x, y=p, mode='lines',
                line=dict(color='green', width=10)
            ),
            go.Surface(z=fc.T, x=s, y=p, colorscale='Blues', showscale=False)
        ])
        return fig


@sh.add_function(dsp, outputs=['fuel_map_model'])
def define_fuel_map_model(fuel_map, full_load_curve):
    """
    Define the fuel map model.

    :param fuel_map:
        Fuel consumption map [RPM, kW, g/s].
    :type fuel_map: dict

    :param full_load_curve:
        Vehicle full load curve.
    :type full_load_curve: function

    :return:
        Fuel map model.
    :rtype: FuelMapModel
    """
    return FuelMapModel(fuel_map, full_load_curve)


def _interp(x, xp, fp):
    x = np.asarray(x).clip(xp[:, 0, None], xp[:, -1, None])
    j = np.maximum(1, np.argmax(x[:, :, None] <= xp[:, None, :], axis=-1))
    i, k = j - 1, np.arange(xp.shape[0])[:, None]
    (x0, dx), (y0, dy) = xp[k, [i, j]], fp[k, [i, j]]
    dx -= x0
    dy -= y0
    with np.errstate(divide='ignore', invalid='ignore'):
        return (x - x0) * np.where(np.isclose(dx, 0), 0, dy / dx) + y0, x


# noinspection PyMissingOrEmptyDocstring,PyArgumentEqualDefault
class HEV:
    def __init__(self, drive_line_efficiencies, motors_efficiencies):

        self.m_eff = np.array(motors_efficiencies)
        m_dl_eff = np.multiply.accumulate(drive_line_efficiencies)
        self.ice_eff = m_dl_eff[-2]
        m_dl_eff = np.insert(m_dl_eff, 1, m_dl_eff[1])
        m_dl_eff = np.insert(m_dl_eff, 0, m_dl_eff[0])
        # Electric assist (dp < 0).
        self.m_ds = m_dl_eff / self.ice_eff
        self.i_ds = np.argsort(self.m_eff * self.m_ds)[::-1]
        self.j_ds = np.argsort(self.i_ds)

        # Charging (dp > 0).
        self.m_ch = m_ch_eff = self.m_ds.copy()
        m_ch_eff[-1] = 1 / m_ch_eff[-1]
        self.i_ch = np.argsort(m_ch_eff / self.m_eff)
        self.j_ch = np.argsort(self.i_ch)

    def delta_ice_power(self, motors_maximums_powers):
        mmp = np.asarray(motors_maximums_powers)
        n, acc = mmp.shape[-1], np.add.accumulate
        mmp = mmp.reshape((mmp.shape[:-1] or (1,)) + (n,))
        dp_ice, p_bat = np.zeros((2, 2 * n + 1) + mmp.shape[:-1])
        dp_ice[:n] = -acc((mmp * self.m_ds).take(self.i_ds, axis=-1).T)[::-1]
        dp_ice[-n:] = acc((mmp * self.m_ch).take(self.i_ch, axis=-1).T)
        p_bat[:n] = -acc((mmp / self.m_eff).take(self.i_ds, axis=-1).T)[::-1]
        p_bat[-n:] = acc((mmp * self.m_eff).take(self.i_ch, axis=-1).T)

        # noinspection PyUnusedLocal
        def battery_power_split(battery_powers, *args):
            p = np.asarray(battery_powers)
            p = p.reshape((p.shape[:-1] or (1,)) + p.shape[-1:])
            b = p < 0
            pm = np.where(b, -p_bat[:n + 1][::-1], p_bat[-1 - n:])
            pm = np.diff(pm + np.minimum(np.abs(p) - pm, 0), axis=0)
            pm = np.where(b, pm[self.j_ds], -pm[self.j_ch])
            return pm

        return dp_ice, p_bat, battery_power_split

    def ice_power(self, motive_power, motors_maximum_powers):
        dp_ice, p_bat, power_split = self.delta_ice_power(motors_maximum_powers)
        p_ice = dp_ice + motive_power / self.ice_eff
        return p_ice, p_bat, power_split

    def parallel(self, motive_powers, motors_maximums_powers,
                 engine_powers_out=None, ice_power_losses=0,
                 battery_power_losses=0):
        pi, pb, bps = self.ice_power(motive_powers, motors_maximums_powers)
        if not isinstance(ice_power_losses, Number) or ice_power_losses != 0:
            pi += np.asarray(ice_power_losses).ravel()[None, :]
        if (not isinstance(battery_power_losses, Number) or
                battery_power_losses != 0):
            pb += np.asarray(battery_power_losses).ravel()[None, :]
        if engine_powers_out is not None:
            pb, pi = _interp(engine_powers_out, pi.T, pb.T)
        return pi, pb, bps

    def ev(self, motive_powers, motors_maximums_powers, battery_power_losses=0):
        return self.parallel(motive_powers, np.pad(
            motors_maximums_powers[:, :-3], ((0, 0), (0, 3)), 'constant'
        ), 0, battery_power_losses=battery_power_losses)

    def serial(self, motive_powers, motors_maximums_powers, engine_powers_out,
               ice_power_losses=0, battery_power_losses=0):
        pi, pb_ev, bps_ev = self.ev(motive_powers, motors_maximums_powers)
        from ..electrics.motors.p4 import calculate_motor_p4_powers_v1 as func
        pw = func(pi.ravel(), self.ice_eff)
        pi, pb, bps = self.parallel(
            pw, np.pad(
                motors_maximums_powers[:, -3:], ((0, 0), (5, 0)), 'constant'
            ), engine_powers_out=engine_powers_out,
            ice_power_losses=ice_power_losses,
            battery_power_losses=battery_power_losses
        )

        def battery_power_split(battery_powers):
            return bps(battery_powers - pb_ev.T) + bps_ev(pb_ev.T)

        return pi, pb + pb_ev, battery_power_split


@sh.add_function(dsp, outputs=['hev_power_model'])
def define_hev_power_model(motors_efficiencies, drive_line_efficiencies):
    """
    Define Hybrid Electric Vehicle power balance model.

    :param motors_efficiencies:
        Electric motors efficiencies vector.
    :type motors_efficiencies: tuple[float]

    :param drive_line_efficiencies:
        Drive line efficiencies vector.
    :type drive_line_efficiencies: tuple[float]

    :return:
        Hybrid Electric Vehicle power balance model.
    :rtype: HEV
    """
    return HEV(drive_line_efficiencies, motors_efficiencies)


@sh.add_function(dsp, outputs=['hybrid_modes'])
def identify_hybrid_modes(
        times, gear_box_speeds_in, engine_speeds_out, idle_engine_speed,
        on_engine, is_serial, has_motor_p2_planetary):
    """
    Identify the hybrid mode status (0: EV, 1: Parallel, 2: Serial).

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param gear_box_speeds_in:
        Gear box speed [RPM].
    :type gear_box_speeds_in: numpy.array

    :param engine_speeds_out:
        Engine speed [RPM].
    :type engine_speeds_out: numpy.array

    :param idle_engine_speed:
        Idle engine speed and its standard deviation [RPM].
    :type idle_engine_speed: (float, float)

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param is_serial:
        Is the vehicle serial hybrid?
    :type is_serial: bool

    :param has_motor_p2_planetary:
        Has the vehicle a motor in planetary P2?
    :type has_motor_p2_planetary: bool

    :return:
        Hybrid mode status (0: EV, 1: Parallel, 2: Serial).
    :rtype: numpy.array
    """
    # noinspection PyProtectedMember
    from ..gear_box.mechanical import _shift
    from ....report import _correlation_coefficient
    if has_motor_p2_planetary:
        return on_engine.astype(int)
    if is_serial:
        return np.where(on_engine, 2, 0)
    mode = on_engine.astype(int)
    es, gbs = engine_speeds_out, gear_box_speeds_in
    b = idle_engine_speed[0] > gbs
    b |= (gbs - idle_engine_speed[1]) > es
    mode[on_engine & b] = 2
    i = np.where(mode == 1)[0]
    mode[i[~co2_utl.get_inliers(gbs[i] / es[i], 3)[0]]] = 2
    mode = co2_utl.median_filter(times, mode, 4)
    mode = co2_utl.clear_fluctuations(times, mode, 4).astype(int)
    mode[~on_engine] = 0
    with np.errstate(divide='ignore', invalid='ignore'):
        for i, j in co2_utl.pairwise(_shift(mode)):
            if mode[i] and times[j - 1] - times[i] < 5:
                if _correlation_coefficient(es[i:j], gbs[i:j]) < .6:
                    mode[i:j] = 3 - mode[i]
    return mode


@sh.add_function(dsp, outputs=['engine_speeds_base'])
def identify_engine_speeds_base(
        times, hybrid_modes, on_engine, engine_speeds_out, gear_box_speeds_in,
        idle_engine_speed):
    """
    Identify the engine speed at hot condition [RPM].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param hybrid_modes:
        Hybrid mode status (0: EV, 1: Parallel, 2: Serial).
    :type hybrid_modes: numpy.array

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param engine_speeds_out:
        Engine speed [RPM].
    :type engine_speeds_out: numpy.array

    :param gear_box_speeds_in:
        Gear box speed [RPM].
    :type gear_box_speeds_in: numpy.array

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :return:
        Engine speed at hot condition [RPM].
    :rtype: numpy.array, float
    """
    b = hybrid_modes == 2
    from .conventional import calculate_engine_speeds_out_hot as func
    speeds = func(gear_box_speeds_in, on_engine, idle_engine_speed)
    speeds[b] = engine_speeds_out[b]
    speeds[b] = co2_utl.median_filter(times, speeds, 5)[b]
    return speeds


@sh.add_function(dsp, outputs=['serial_motor_maximum_power_function'])
def define_serial_motor_maximum_power_function(
        motor_p2_planetary_maximum_power_function, planetary_ratio,
        motor_p1_maximum_power_function, motor_p0_maximum_power_function,
        motor_p2_planetary_speed_ratio, motor_p1_speed_ratio,
        motor_p0_speed_ratio):
    """
    Define serial motor maximum power function.

    :param motor_p2_planetary_maximum_power_function:
        Maximum power function of planetary motor P2.
    :type motor_p2_planetary_maximum_power_function: function

    :param planetary_ratio:
        Fundamental planetary speed ratio [-].
    :type planetary_ratio: float

    :param motor_p2_planetary_speed_ratio:
        Ratio between planetary motor P2 speed and planetary speed [-].
    :type motor_p2_planetary_speed_ratio: float

    :param motor_p1_maximum_power_function:
        Maximum power function of motor P1.
    :type motor_p1_maximum_power_function: function

    :param motor_p1_speed_ratio:
        Ratio between motor P1 speed and engine speed [-].
    :type motor_p1_speed_ratio: float

    :param motor_p0_maximum_power_function:
        Maximum power function of motor P0.
    :type motor_p0_maximum_power_function: function

    :param motor_p0_speed_ratio:
        Ratio between motor P0 speed and engine speed [-].
    :type motor_p0_speed_ratio: float

    :return:
        Serial motor maximum power function.
    :rtype: function
    """
    planet_f = motor_p2_planetary_maximum_power_function
    from ..electrics.motors.planet import calculate_planetary_speeds_in as func

    # noinspection PyMissingOrEmptyDocstring
    def calculate_serial_motor_maximum_power(engine_speed, final_drive_speed):
        es = np.atleast_1d(engine_speed)
        ps = func(engine_speed, final_drive_speed, planetary_ratio)
        # noinspection PyArgumentEqualDefault
        return np.pad(
            np.column_stack((
                planet_f(ps * motor_p2_planetary_speed_ratio),
                motor_p1_maximum_power_function(es * motor_p1_speed_ratio),
                motor_p0_maximum_power_function(es * motor_p0_speed_ratio)
            )), ((0, 0), (5, 0)), 'constant'
        )

    return calculate_serial_motor_maximum_power


dsp.add_data('min_engine_on_speed', dfl.values.min_engine_on_speed)


@sh.add_function(dsp, outputs=['engine_power_losses_function'])
def define_engine_power_losses_function(
        engine_moment_inertia, auxiliaries_torque_loss, auxiliaries_power_loss,
        min_engine_on_speed):
    """
    Define engine power losses function.

    :param engine_moment_inertia:
        Engine moment of inertia [kg*m2].
    :type engine_moment_inertia: float

    :param auxiliaries_torque_loss:
        Constant torque loss due to engine auxiliaries [N*m].
    :type auxiliaries_torque_loss: float

    :param auxiliaries_power_loss:
        Constant power loss due to engine auxiliaries [kW].
    :type auxiliaries_power_loss: float

    :param min_engine_on_speed:
        Minimum engine speed to consider the engine to be on [RPM].
    :type min_engine_on_speed: float

    :return:
        Engine power losses function.
    :rtype: function
    """
    from ..engine import (
        calculate_auxiliaries_power_losses as aux_p,
        calculate_auxiliaries_torque_losses as aux_t,
        calculate_engine_inertia_powers_losses as ine_p
    )

    # noinspection PyMissingOrEmptyDocstring
    def engine_power_losses_function(times, engine_speeds, inertia=True):
        p = aux_p(
            aux_t(times, auxiliaries_torque_loss), engine_speeds,
            np.ones_like(engine_speeds, bool), auxiliaries_power_loss
        )
        if inertia:
            p += ine_p(times, engine_speeds, engine_moment_inertia)

        return np.where(engine_speeds < min_engine_on_speed, 0, p)

    return engine_power_losses_function


# noinspection PyMissingOrEmptyDocstring
class EMS:
    def __init__(self, has_motor_p2_planetary, is_serial, battery_model,
                 hev_power_model, fuel_map_model,
                 serial_motor_maximum_power_function,
                 engine_power_losses_function, s_ch=None, s_ds=None):
        self.battery_model = battery_model
        self.hev_power_model = hev_power_model
        self.fuel_map_model = fuel_map_model
        self.s_ch, self.s_ds = s_ch, s_ds
        self.battery_fuel_mode = 'current'
        self._battery_power = None
        self.serial_motor_maximum_power = serial_motor_maximum_power_function
        self.engine_power_losses = engine_power_losses_function
        self.has_motor_p2_planetary = has_motor_p2_planetary
        self.is_serial = is_serial

    def set_virtual(self, motors_maximum_powers):
        from scipy.interpolate import UnivariateSpline as Spline
        p, pb = self.hev_power_model.delta_ice_power(motors_maximum_powers)[:-1]
        pb, i = np.unique(-pb, return_index=True)
        self._battery_power = Spline(pb, np.abs(p.ravel()[i]), k=1, s=0)

    # noinspection PyUnresolvedReferences
    def fit(self, hybrid_modes, times, motive_powers, motors_maximums_powers,
            engine_powers_out, engine_speeds_out, ice_power_losses=None):
        pl = ice_power_losses
        if pl is None:
            pl = self.engine_power_losses(times, engine_speeds_out)
        hev, fc = self.hev_power_model, self.fuel_map_model
        pi = engine_powers_out[:, None] + [-dfl.EPS, dfl.EPS]
        pb = np.where(hybrid_modes[:, None] == 1, *(func(
            motive_powers, motors_maximums_powers, pi, ice_power_losses=pl
        )[1] for func in (hev.parallel, hev.serial)))
        b = hybrid_modes.astype(bool)
        pi, bc = pi[b], self.battery_model.currents(pb[b])
        s = -np.diff(fc(engine_speeds_out[b, None], pi), axis=1)
        with np.errstate(divide='ignore', invalid='ignore'):
            s /= np.diff(bc, axis=1)
        b = np.isfinite(s)
        b, s = bc[b.ravel()].mean(1) >= 0, s[b]
        self.s_ch, self.s_ds = np.median(s[b]), np.median(s[~b])
        return self

    def battery_fuel(self, battery_currents):
        bc = battery_currents
        if self.battery_fuel_mode == 'current':
            return np.where(bc >= 0, self.s_ch, self.s_ds) * bc
        pi = self._battery_power(self.battery_model.powers(bc))
        fc = self.fuel_map_model(self.fuel_map_model.power_speed(pi), pi)
        fc -= self.fuel_map_model.idle_fc
        fc *= -np.sign(bc)
        return fc

    def electric(self, motive_powers, motors_maximums_powers,
                 final_drive_speeds_in, battery_power_losses=0):
        res, hev = {}, self.hev_power_model
        mmp = self.serial_motor_maximum_power(
            np.zeros_like(motive_powers), final_drive_speeds_in
        )
        mmp[:, :-3] = motors_maximums_powers[:, :-3]
        pi, pb, bps = hev.ice_power(motive_powers, mmp)
        res['power_bat'], res['power_ice'] = pb, pi = _interp(0, pi.T, pb.T)
        pb += np.atleast_2d(battery_power_losses).T
        res['current_bat'] = bc = self.battery_model.currents(pb)
        res['fc_bat'] = res['fc_eq'] = self.battery_fuel(bc)
        res['fc_ice'] = res['speed_ice'] = np.zeros_like(bc)
        res['battery_power_split'] = bps
        return res

    def parallel(self, times, motive_powers, motors_maximums_powers,
                 engine_speeds_out, ice_power_losses=None,
                 battery_power_losses=0, opt=True):
        hev, fc = self.hev_power_model, self.fuel_map_model
        pi, pb, bps = hev.ice_power(motive_powers, motors_maximums_powers)
        if ice_power_losses is None:
            ice_power_losses = self.engine_power_losses(
                times, engine_speeds_out
            )
        pi += np.atleast_2d(ice_power_losses)
        ep = np.minimum(fc.full_load_curve(engine_speeds_out), pi[-1])
        ep = np.linspace(0, 1, 200) * ep[:, None]
        pb, pi = _interp(ep, pi.T, pb.T)
        es = np.atleast_2d(engine_speeds_out).T
        fc_ice = fc(es, pi)
        pb += np.atleast_2d(battery_power_losses).T
        bc = self.battery_model.currents(pb)
        fc_bat = self.battery_fuel(bc)
        fc_eq = fc_ice + fc_bat
        res = dict(
            fc_eq=fc_eq, fc_ice=fc_ice, fc_bat=fc_bat, power_bat=pb,
            power_ice=pi, current_bat=bc, speed_ice=es, battery_power_split=bps
        )
        if opt:
            i, j = np.arange(fc_eq.shape[0]), np.nanargmin(fc_eq, axis=1)
            res = self.min(res, i, j)
        return res

    def serial(self, times, motive_powers, final_drive_speeds_in,
               motors_maximums_powers, engine_speeds_out=None,
               ice_power_losses=None, battery_power_losses=0, opt=True):
        hev, fc, n = self.hev_power_model, self.fuel_map_model, 200
        # noinspection PyArgumentEqualDefault
        pi, pb, bps_ev = hev.ice_power(motive_powers, np.pad(
            motors_maximums_powers[:, :-3], ((0, 0), (0, 3)), 'constant'
        ))
        (pb_ev, pi), pl = _interp(0, pi.T, pb.T), ice_power_losses
        pb_ev, fds = pb_ev.ravel(), final_drive_speeds_in[:, None]
        from ..electrics.motors.p4 import calculate_motor_p4_powers_v1 as func
        pw = func(pi.ravel(), hev.ice_eff)
        del pi
        if engine_speeds_out is None:
            # noinspection PyProtectedMember
            es = self.fuel_map_model.speed_power._data[0]
            es = np.linspace(es[0], es[-1], n)[:, None]
            if not hasattr(pl, 'shape'):
                pl = np.ones_like(motive_powers) * (pl or 0)

            es, pl = np.meshgrid(es, pl, indexing='ij')
            if ice_power_losses is None:
                pl = self.engine_power_losses(times, es, inertia=False)
            es, pl = es.ravel()[:, None], pl.ravel()
            fds = np.tile(fds, n).ravel()[:, None]
            pw = np.tile(pw[:, None], n).ravel()
        else:
            es = engine_speeds_out[:, None]
            if pl is None:
                pl = self.engine_power_losses(times, engine_speeds_out)
        mmp = self.serial_motor_maximum_power(es, fds)
        pi, pb, bps = hev.ice_power(pw, mmp)
        pi += np.atleast_2d(pl)
        pb, pi = _interp(fc.speed_power(es), pi.T, pb.T)
        if engine_speeds_out is None:
            pb, pi = pb.reshape(n, -1).T, pi.reshape(n, -1).T
            es = es.reshape(n, -1).T
        pb += np.atleast_2d(pb_ev + battery_power_losses).T
        fc_ice = fc(es, pi)
        bc = self.battery_model.currents(pb)
        fc_bat = self.battery_fuel(bc)
        fc_eq = fc_ice + fc_bat

        def battery_power_split(
                battery_powers, engine_speeds, final_drive_speeds):
            return hev.ice_power(
                0, self.serial_motor_maximum_power(
                    engine_speeds, final_drive_speeds
                )
            )[-1](battery_powers - battery_power_losses - pb_ev) + bps_ev(pb_ev)

        res = dict(
            fc_eq=fc_eq, fc_ice=fc_ice, fc_bat=fc_bat, power_bat=pb,
            power_ice=pi, current_bat=bc, speed_ice=es,
            battery_power_split=battery_power_split
        )
        if engine_speeds_out is None and opt:
            with np.errstate(divide='ignore', invalid='ignore'):
                eff, b = pi / fc_ice, (bc < 0) | (pi < 0)
            b[np.all(np.isnan(eff) | b, 1)] = False
            eff[b] = np.nan
            i, j = np.arange(pi.shape[0]), np.nanargmax(eff, 1)
            res = self.min(res, i, j)
            res['speed_ice'] = res['speed_ice'][i, j, None]
        return res

    @staticmethod
    def min(res, i, j):
        res, keys = res.copy(), set(res) - {'battery_power_split', 'speed_ice'}
        for k in keys:
            res[k] = res[k][i, j, None]
        return res

    def __call__(self, times, motive_powers, final_drive_speeds_in,
                 motors_maximums_powers, gear_box_speeds_in,
                 engine_speeds_parallel, idle_engine_speed,
                 battery_power_losses=0):
        e = self.electric(
            motive_powers, motors_maximums_powers, final_drive_speeds_in
        )
        p = self.parallel(
            times, motive_powers, motors_maximums_powers,
            engine_speeds_parallel, battery_power_losses=battery_power_losses
        )
        s = self.serial(
            times, motive_powers, final_drive_speeds_in, motors_maximums_powers,
            battery_power_losses=battery_power_losses
        )
        if self.is_serial:
            mode = np.zeros_like(s['fc_eq'], int) + 2
            be_serial = np.zeros_like(times, bool)
        else:
            # noinspection PyUnresolvedReferences
            mode = (s['fc_eq'] < p['fc_eq']).astype(int) + 1
            mode[(s['current_bat'] < 0) & (p['current_bat'] >= 0)] = 1
            mode[(s['current_bat'] >= 0) & (p['current_bat'] < 0)] = 2
            mode[gear_box_speeds_in < -np.diff(idle_engine_speed)] = 2
            be_serial = e['power_ice'].ravel() > dfl.EPS
            be_serial &= motive_powers > 0.01
            mode[be_serial] = 1
        if self.has_motor_p2_planetary:
            mode[mode == 2] = 1
        return dict(
            hybrid_modes=mode, serial=s, parallel=p, electric=e,
            force_on_engine=be_serial
        )


@sh.add_function(dsp, outputs=['ecms_s'])
def calibrate_ems_model(
        drive_battery_model, hev_power_model, fuel_map_model, hybrid_modes,
        serial_motor_maximum_power_function, motive_powers,
        motors_maximums_powers, engine_powers_out, engine_speeds_out, times,
        engine_power_losses_function, is_serial, has_motor_p2_planetary):
    """
    Calibrate Energy Management Strategy model.

    :param drive_battery_model:
        Drive battery current model.
    :type drive_battery_model: DriveBatteryModel

    :param hev_power_model:
        Hybrid Electric Vehicle power balance model.
    :type hev_power_model: HEV

    :param fuel_map_model:
        Fuel map model.
    :type fuel_map_model: FuelMapModel

    :param hybrid_modes:
        Hybrid mode status (0: EV, 1: Parallel, 2: Serial).
    :type hybrid_modes: numpy.array

    :param serial_motor_maximum_power_function:
        Serial motor maximum power function.
    :type serial_motor_maximum_power_function: function

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :param motors_maximums_powers:
        Maximum powers of electric motors [kW].
    :type motors_maximums_powers: numpy.array

    :param engine_powers_out:
        Engine power vector [kW].
    :type engine_powers_out: numpy.array

    :param engine_speeds_out:
        Engine speed [RPM].
    :type engine_speeds_out: numpy.array

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param engine_power_losses_function:
        Engine power losses function.
    :type engine_power_losses_function: function

    :param is_serial:
        Is the vehicle serial hybrid?
    :type is_serial: bool

    :param has_motor_p2_planetary:
        Has the vehicle a motor in planetary P2?
    :type has_motor_p2_planetary: bool

    :return:
        Equivalent Consumption Minimization Strategy params.
    :rtype: tuple[float]
    """
    model = EMS(
        has_motor_p2_planetary, is_serial, drive_battery_model, hev_power_model,
        fuel_map_model, serial_motor_maximum_power_function,
        engine_power_losses_function
    ).fit(
        hybrid_modes, times, motive_powers, motors_maximums_powers,
        engine_powers_out, engine_speeds_out
    )
    return model.s_ch, model.s_ds


@sh.add_function(dsp, outputs=['ems_model'])
def define_ems_model(
        has_motor_p2_planetary, is_serial, drive_battery_model, hev_power_model,
        fuel_map_model, serial_motor_maximum_power_function,
        engine_power_losses_function, ecms_s):
    """
    Define Energy Management Strategy model.

    :param has_motor_p2_planetary:
        Has the vehicle a motor in planetary P2?
    :type has_motor_p2_planetary: bool

    :param is_serial:
        Is the vehicle serial hybrid?
    :type is_serial: bool

    :param drive_battery_model:
        Drive battery current model.
    :type drive_battery_model: DriveBatteryModel

    :param hev_power_model:
        Hybrid Electric Vehicle power balance model.
    :type hev_power_model: HEV

    :param fuel_map_model:
        Fuel map model.
    :type fuel_map_model: FuelMapModel

    :param serial_motor_maximum_power_function:
        Serial motor maximum power function.
    :type serial_motor_maximum_power_function: function

    :param engine_power_losses_function:
        Engine power losses function.
    :type engine_power_losses_function: function

    :param ecms_s:
        Equivalent Consumption Minimization Strategy params.
    :type ecms_s: tuple[float]

    :return:
        Energy Management Strategy model.
    :rtype: EMS
    """
    return EMS(
        has_motor_p2_planetary, is_serial, drive_battery_model, hev_power_model,
        fuel_map_model, serial_motor_maximum_power_function,
        engine_power_losses_function, s_ch=ecms_s[0], s_ds=ecms_s[1]
    )


@sh.add_function(dsp, outputs=['ems_data'])
def calculate_ems_data(
        ems_model, times, motive_powers, final_drive_speeds_in,
        motors_maximums_powers, gear_box_speeds_in, engine_speeds_parallel,
        idle_engine_speed):
    """
    Calculate EMS decision data.

    :param ems_model:
        Energy Management Strategy model.
    :type ems_model: EMS

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :param final_drive_speeds_in:
        Final drive speed in [RPM].
    :type final_drive_speeds_in: numpy.array

    :param motors_maximums_powers:
        Maximum powers of electric motors [kW].
    :type motors_maximums_powers: numpy.array

    :param gear_box_speeds_in:
        Gear box speed [RPM].
    :type gear_box_speeds_in: numpy.array

    :param engine_speeds_parallel:
        Hypothetical engine speed in parallel mode [RPM].
    :type engine_speeds_parallel: numpy.array

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :return:
        EMS decision data.
    :rtype: dict
    """
    return ems_model(
        times, motive_powers, final_drive_speeds_in, motors_maximums_powers,
        gear_box_speeds_in, engine_speeds_parallel, idle_engine_speed
    )


def _filter_warm_up(
        times, warm_up, on_engine, engine_thermostat_temperature, is_cycle_hot,
        engine_temperatures, hybrid_modes=None):
    if hybrid_modes is None:
        hybrid_modes = np.ones_like(times) * 2
    indices, temp = co2_utl.index_phases(warm_up), engine_temperatures
    b = np.diff(temp[indices], axis=1) > 4
    b &= temp[indices[:, 0], None] < (engine_thermostat_temperature - 10)
    b &= np.diff(times[indices], axis=1) > 5
    b &= np.apply_along_axis(
        lambda a: np.in1d(2, hybrid_modes[slice(*a)]), 1, indices
    )
    t_cool = dfl.functions.identify_after_treatment_warm_up_phases.cooling_time
    t0 = is_cycle_hot and times[0] + t_cool or 0
    warming_phases = np.zeros_like(on_engine, bool)
    for i, j in co2_utl.index_phases(on_engine):
        warming_phases[i:j + 1] = times[i] >= t0
        t0 = times[j] + t_cool

    warm_up = np.zeros_like(warm_up, bool)
    for i, j in indices[b.ravel()]:
        if warming_phases[i]:
            while i and warming_phases.take(i - 1, mode='clip'):
                i -= 1
            while hybrid_modes[j] == 1 and j > i:
                j -= 1
            warm_up[i:j + 1] = True
    return warm_up


@sh.add_function(
    dsp, inputs_kwargs=True, outputs=['after_treatment_warm_up_phases']
)
def identify_after_treatment_warm_up_phases(
        times, engine_powers_out, engine_temperatures, on_engine,
        engine_thermostat_temperature, ems_data, hybrid_modes,
        is_cycle_hot=False):
    """
    Identifies after treatment warm up phase.

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param engine_powers_out:
        Engine power vector [kW].
    :type engine_powers_out: numpy.array

    :param engine_temperatures:
        Engine coolant temperature vector [°C].
    :type engine_temperatures: numpy.array

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param ems_data:
        EMS decision data.
    :type ems_data: dict

    :param hybrid_modes:
        Hybrid mode status (0: EV, 1: Parallel, 2: Serial).
    :type hybrid_modes: numpy.array

    :param engine_thermostat_temperature:
        Engine thermostat temperature [°C].
    :type engine_thermostat_temperature: float

    :param is_cycle_hot:
        Is an hot cycle?
    :type is_cycle_hot: bool

    :return:
        After treatment warm up phase.
    :rtype: numpy.array
    """
    warm_up = np.zeros_like(times)
    if not on_engine.any():
        return warm_up.astype(bool)
    i = np.where(on_engine)[0]
    p = np.choose(ems_data['hybrid_modes'].ravel() - 1, [
        ems_data[k]['power_ice'].ravel() for k in ('parallel', 'serial')
    ])[i]
    with np.errstate(divide='ignore', invalid='ignore'):
        # noinspection PyUnresolvedReferences
        warm_up[i[~co2_utl.get_inliers(
            engine_powers_out[i] / p, 2, np.nanmedian, co2_utl.mad
        )[0]]] = 1
    warm_up = co2_utl.median_filter(times, warm_up, 5)
    warm_up = co2_utl.clear_fluctuations(times, warm_up, 5).astype(bool)
    if not warm_up.any():
        return warm_up
    return _filter_warm_up(
        times, warm_up, on_engine, engine_thermostat_temperature, is_cycle_hot,
        engine_temperatures, hybrid_modes
    )


@sh.add_function(dsp, outputs=['force_on_engine'])
def get_force_on_engine(ems_data):
    """
    Returns the phases when engine is on because parallel mode is forced [-].

    :param ems_data:
        EMS decision data.
    :type ems_data: dict

    :return:
        Phases when engine is on because parallel mode is forced [-].
    :rtype: numpy.array
    """
    return ems_data['force_on_engine']


# noinspection PyMissingOrEmptyDocstring
class StartStopHybrid:
    def __init__(self, ems_model, dcdc_converter_efficiency, starter_model,
                 params=None):
        self.params = params
        self.ems_model = ems_model
        self.dcdc_converter_efficiency = dcdc_converter_efficiency
        self.starter_model = starter_model

    def fit(self, ems_data, on_engine, drive_battery_state_of_charges,
            after_treatment_warm_up_phases):
        import lmfit
        k = np.where(~on_engine, *self.reference(ems_data)['k_reference'].T)

        # Filter data.
        b = ~after_treatment_warm_up_phases & ~ems_data['force_on_engine']
        b &= np.isfinite(k)
        s = np.where(on_engine[b], 1, -1)
        soc = drive_battery_state_of_charges[b]

        def _(x):
            x = x.valuesdict()
            time = x['starter_time']
            return np.float32(np.maximum(0, s * (self._k(x, soc) - np.where(
                ~on_engine, *self.reference(ems_data, time)['k_reference'].T
            )[b].T)).sum())

        p = lmfit.Parameters()
        starter_time = self.starter_model.time
        p.add('starter_time', 1, min=starter_time, max=starter_time * 10)
        p.add('k0', 0, min=0)
        p.add('soc0', 0, min=0, max=100)
        p.add('alpha', 0, min=0)
        p.add('beta', 0, min=0)

        with co2_utl.numpy_random_seed(0):
            # noinspection PyUnresolvedReferences
            self.params = lmfit.minimize(_, p, 'ampgo').params.valuesdict()
        return self

    def penalties(self, data, starter_time):
        eff, res = self.dcdc_converter_efficiency, {}
        c = np.array([eff, -1 / eff]) / starter_time
        pb = self.starter_model(data['speed_ice'])[None, :] * c[:, None, None]
        bc = self.ems_model.battery_model.currents(pb)
        return sh.combine_dicts(data, base=dict(
            power_stop=pb[0], power_start=pb[1], current_stop=bc[0],
            current_start=bc[1]
        ))

    def reference(self, ems_data, starter_time=None):
        st = starter_time or self.params and self.params['starter_time']
        st = st or self.starter_model.time
        e, res = ems_data['electric'], ems_data.copy()
        p = res['parallel'] = self.penalties(ems_data['parallel'], st)
        s = res['serial'] = self.penalties(ems_data['serial'], st)
        with np.errstate(divide='ignore', invalid='ignore'):
            k = 'current_bat'
            c_ser = np.column_stack((s['current_start'], -s['current_stop']))
            k_ser = ((s[k] - e[k] + c_ser) / s['fc_ice']).T
            c_par = np.column_stack((p['current_start'], -p['current_stop']))
            k_par = ((p[k] - e[k] + c_par) / p['fc_ice']).T
            res['k_reference'] = np.choose(
                ems_data['hybrid_modes'] - 1, [k_par.T, k_ser.T]
            )
        return res

    @staticmethod
    def _k(p, soc):
        dsoc = soc - p['soc0']
        return dsoc * (p['alpha'] + p['beta'] * dsoc ** 2) + p['k0']

    def __call__(self, drive_battery_state_of_charges):
        return self._k(self.params, drive_battery_state_of_charges)


@sh.add_function(dsp, outputs=['start_stop_hybrid_params'])
def calibrate_start_stop_hybrid_params(
        ems_model, dcdc_converter_efficiency, starter_model, ems_data,
        on_engine, drive_battery_state_of_charges,
        after_treatment_warm_up_phases):
    """
    Calibrate start stop model for hybrid electric vehicles.

    :param ems_model:
        Energy Management Strategy model.
    :type ems_model: EMS

    :param starter_model:
        Starter model.
    :type starter_model: StarterModel

    :param dcdc_converter_efficiency:
        DC/DC converter efficiency [-].
    :type dcdc_converter_efficiency: float

    :param ems_data:
        EMS decision data.
    :type ems_data: dict

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param drive_battery_state_of_charges:
        State of charge of the drive battery [%].
    :type drive_battery_state_of_charges: numpy.array

    :param after_treatment_warm_up_phases:
        Phases when engine is affected by the after treatment warm up [-].
    :type after_treatment_warm_up_phases: numpy.array

    :return:
        Params of start stop model for hybrid electric vehicles.
    :rtype: dict
    """
    return StartStopHybrid(
        ems_model, dcdc_converter_efficiency, starter_model).fit(
        ems_data, on_engine, drive_battery_state_of_charges,
        after_treatment_warm_up_phases
    ).params


@sh.add_function(dsp, outputs=['start_stop_hybrid'])
def define_start_stop_hybrid(
        ems_model, dcdc_converter_efficiency, starter_model,
        start_stop_hybrid_params):
    """
    Defines start stop model for hybrid electric vehicles.

    :param ems_model:
        Energy Management Strategy model.
    :type ems_model: EMS

    :param starter_model:
        Starter model.
    :type starter_model: StarterModel

    :param dcdc_converter_efficiency:
        DC/DC converter efficiency [-].
    :type dcdc_converter_efficiency: float

    :param start_stop_hybrid_params:
        Params of start stop model for hybrid electric vehicles.
    :type start_stop_hybrid_params: dict

    :return:
        Start stop model for hybrid electric vehicles.
    :rtype: StartStopHybrid
    """
    return StartStopHybrid(
        ems_model, dcdc_converter_efficiency, starter_model,
        params=start_stop_hybrid_params
    )


@sh.add_function(dsp, outputs=['hybrid_modes'])
def calculate_hybrid_modes(ems_data, on_engine, after_treatment_warm_up_phases):
    """
    Calculate the hybrid mode status (0: EV, 1: Parallel, 2: Serial).

    :param ems_data:
        EMS decision data.
    :type ems_data: dict

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :param after_treatment_warm_up_phases:
        Phases when engine is affected by the after treatment warm up [-].
    :type after_treatment_warm_up_phases: numpy.array

    :return:
        Hybrid mode status (0: EV, 1: Parallel, 2: Serial).
    :rtype: numpy.array
    """
    modes = ems_data['hybrid_modes'].ravel().copy()
    modes[after_treatment_warm_up_phases] = 2
    modes[~on_engine] = 0
    return modes


@sh.add_function(dsp, outputs=['hybrid_modes'])
def predict_hybrid_modes(
        start_stop_hybrid, ems_data, drive_battery_model, times, motive_powers,
        accelerations, after_treatment_warm_up_duration,
        after_treatment_cooling_duration, start_stop_activation_time,
        min_time_engine_on_after_start, is_cycle_hot):
    """
    Predicts the hybrid mode status (0: EV, 1: Parallel, 2: Serial).

    :param start_stop_hybrid:
        Start stop model for hybrid electric vehicles.
    :type start_stop_hybrid: StartStopHybrid

    :param ems_data:
        EMS decision data.
    :type ems_data: dict

    :param drive_battery_model:
        Drive battery current model.
    :type drive_battery_model: DriveBatteryModel

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :param accelerations:
        Acceleration [m/s2].
    :type accelerations: numpy.array

    :param after_treatment_warm_up_duration:
        After treatment warm up duration [s].
    :type after_treatment_warm_up_duration: float

    :param after_treatment_cooling_duration:
        After treatment cooling duration [s].
    :type after_treatment_cooling_duration: float

    :param min_time_engine_on_after_start:
        Minimum time of engine on after a start [s].
    :type min_time_engine_on_after_start: float

    :param start_stop_activation_time:
        Start-stop activation time threshold [s].
    :type start_stop_activation_time: float

    :param is_cycle_hot:
        Is an hot cycle?
    :type is_cycle_hot: bool

    :return:
        Hybrid mode status (0: EV, 1: Parallel, 2: Serial).
    :rtype: numpy.array
    """
    r = start_stop_hybrid.reference(ems_data)

    ele, par, ser = r['electric'], r['parallel'], r['serial']
    current_bat = {
        i: v['current_bat'].ravel() for i, v in enumerate((ele, par, ser, ele))
    }
    starter_bat = {
        i: v['power_start'].ravel() for i, v in enumerate((par, ser, ser), 1)
    }
    starter_bat[0] = np.where(
        r['hybrid_modes'] == 1, *(v['power_stop'].T for v in (par, ser))
    ).ravel()
    from ..electrics.motors.starter import calculate_starter_currents as func
    nom_volt = drive_battery_model.service.nominal_voltage
    starter_bat = {k: func(v, nom_volt) for k, v in starter_bat.items()}

    # noinspection PyUnresolvedReferences
    hybrid_modes = r['force_on_engine'].astype(int)
    drive_battery_model.reset()
    soc, t0 = drive_battery_model.init_soc, start_stop_activation_time
    it = enumerate(zip(
        times, motive_powers, accelerations, hybrid_modes, r['k_reference'],
        r['hybrid_modes'].ravel()
    ))
    t_after_treatment = times[0]
    if is_cycle_hot:
        t_after_treatment += after_treatment_cooling_duration
    for i, (t, motive_power, acc, mode, k_ref, mode_ref) in it:
        pre_mode = hybrid_modes.take(i - 1, mode='clip')
        j = int(bool(pre_mode))
        if not mode:
            if t < t0:
                mode = pre_mode
            elif k_ref[j] > start_stop_hybrid(soc):
                mode = mode_ref
        starter_curr = 0
        if bool(pre_mode) ^ bool(mode) and i:
            if mode:
                t0 = t + min_time_engine_on_after_start
                if t >= t_after_treatment:
                    if not hybrid_modes[i]:
                        mode = 3
                    j = np.searchsorted(
                        times, t + after_treatment_warm_up_duration
                    ) + 1
                    hybrid_modes[i:j][hybrid_modes[i:j] == 0] = 3
            else:
                t_after_treatment = t + after_treatment_cooling_duration
            starter_curr = starter_bat[mode][i]
        soc = drive_battery_model(
            current_bat[mode][i], t, motive_power, acc, bool(mode), starter_curr
        )
        hybrid_modes[i] = mode
    return np.minimum(hybrid_modes, 2)


@sh.add_function(dsp, outputs=['on_engine'])
def identify_on_engine(hybrid_modes):
    """
    Identifies if the engine is on [-].

    :param hybrid_modes:
        Hybrid mode status (0: EV, 1: Parallel, 2: Serial).
    :type hybrid_modes: numpy.array

    :return:
        If the engine is on [-].
    :rtype: numpy.array
    """
    return hybrid_modes.astype(bool)


@sh.add_function(dsp, outputs=['engine_speeds_out_hot'])
def identify_engine_speeds_out_hot(
        engine_speeds_base, after_treatment_warm_up_phases, idle_engine_speed):
    """
    Predicts the engine speed at hot condition [RPM].

    :param engine_speeds_base:
        Base engine speed (i.e., without clutch/TC effect) [RPM].
    :type engine_speeds_base: numpy.array

    :param after_treatment_warm_up_phases:
        Phases when engine is affected by the after treatment warm up [-].
    :type after_treatment_warm_up_phases: numpy.array

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :return:
        Engine speed at hot condition [RPM].
    :rtype: numpy.array
    """
    return np.where(
        after_treatment_warm_up_phases, idle_engine_speed[0], engine_speeds_base
    )


@sh.add_function(dsp, outputs=['engine_speeds_out_hot'], weight=1)
def predict_engine_speeds_out_hot(
        ems_data, hybrid_modes, after_treatment_warm_up_phases,
        idle_engine_speed):
    """
    Predicts the engine speed at hot condition [RPM].

    :param ems_data:
        EMS decision data.
    :type ems_data: dict

    :param hybrid_modes:
        Hybrid mode status (0: EV, 1: Parallel, 2: Serial).
    :type hybrid_modes: numpy.array

    :param after_treatment_warm_up_phases:
        Phases when engine is affected by the after treatment warm up [-].
    :type after_treatment_warm_up_phases: numpy.array

    :param idle_engine_speed:
        Engine speed idle median and std [RPM].
    :type idle_engine_speed: (float, float)

    :return:
        Engine speed at hot condition [RPM].
    :rtype: numpy.array
    """
    it = ems_data['electric'], ems_data['parallel'], ems_data['serial']
    speeds = np.choose(hybrid_modes, [d['speed_ice'].ravel() for d in it])
    phases = after_treatment_warm_up_phases
    speeds[phases & (hybrid_modes == 2)] = idle_engine_speed[0]
    return speeds


@sh.add_function(dsp, outputs=[
    'motor_p4_front_electric_powers', 'motor_p4_rear_electric_powers',
    'motor_p3_front_electric_powers', 'motor_p3_rear_electric_powers',
    'motor_p2_electric_powers', 'motor_p2_planetary_electric_powers',
    'motor_p1_electric_powers', 'motor_p0_electric_powers'
])
def predict_motors_electric_powers(
        ems_data, after_treatment_warm_up_phases, hybrid_modes,
        engine_speeds_out_hot, final_drive_speeds_in):
    """
    Predicts motors electric power split [kW].

    :param ems_data:
        EMS decision data.
    :type ems_data: dict

    :param after_treatment_warm_up_phases:
        Phases when engine is affected by the after treatment warm up [-].
    :type after_treatment_warm_up_phases: numpy.array

    :param hybrid_modes:
        Hybrid mode status (0: EV, 1: Parallel, 2: Serial).
    :type hybrid_modes: numpy.array

    :param engine_speeds_out_hot:
        Engine speed at hot condition [RPM].
    :type engine_speeds_out_hot: numpy.array

    :param final_drive_speeds_in:
        Final drive speed in [RPM].
    :type final_drive_speeds_in: numpy.array

    :return:
        Motors electric powers [kW].
    :rtype: tuple[numpy.array]
    """
    mode = np.where(
        after_treatment_warm_up_phases & (hybrid_modes == 2), 0, hybrid_modes
    )
    return np.choose(mode, [
        d['battery_power_split'](
            d['power_bat'].ravel(), engine_speeds_out_hot, final_drive_speeds_in
        ) for d in (ems_data[k] for k in ('electric', 'parallel', 'serial'))
    ])


dsp.add_data('is_serial', dfl.values.is_serial)


@sh.add_function(dsp, outputs=[
    'motor_p4_front_electric_powers', 'motor_p4_rear_electric_powers',
    'motor_p3_front_electric_powers', 'motor_p3_rear_electric_powers',
    'motor_p2_electric_powers', 'motor_p2_planetary_electric_powers',
    'motor_p1_electric_powers', 'motor_p0_electric_powers'
])
def identify_motors_electric_powers(
        hev_power_model, hybrid_modes, motive_powers, motors_maximums_powers,
        motors_electric_powers):
    """
    Identify motors electric power split [kW].

    :param hev_power_model:
        Hybrid Electric Vehicle power balance model.
    :type hev_power_model: HEV

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :param hybrid_modes:
        Hybrid mode status (0: EV, 1: Parallel, 2: Serial).
    :type hybrid_modes: numpy.array

    :param motors_maximums_powers:
        Maximum powers of electric motors [kW].
    :type motors_maximums_powers: numpy.array

    :param motors_electric_powers:
        Cumulative motors electric power [kW].
    :type motors_electric_powers: numpy.array

    :return:
        Motors electric powers [kW].
    :rtype: tuple[numpy.array]
    """
    hev, pb = hev_power_model, -motors_electric_powers[None, :]
    return np.choose(hybrid_modes, [
        func(motive_powers, motors_maximums_powers, 0)[-1](pb)
        for func in (hev.ev, hev.parallel, hev.serial)
    ])


@sh.add_function(dsp, outputs=[
    'motor_p4_front_electric_powers', 'motor_p4_rear_electric_powers',
    'motor_p3_front_electric_powers', 'motor_p3_rear_electric_powers',
    'motor_p2_electric_powers', 'motor_p2_planetary_electric_powers',
    'motor_p1_electric_powers', 'motor_p0_electric_powers'
], weight=sh.inf(1, 0))
def identify_motors_electric_powers_v1(
        final_drive_mean_efficiency, gear_box_mean_efficiency_guess,
        planetary_mean_efficiency, belt_mean_efficiency, motors_efficiencies,
        hybrid_modes, motive_powers, motors_maximums_powers,
        motors_electric_powers, has_motor_p2_planetary):
    """
    Identify motors electric power split [kW].

    :param final_drive_mean_efficiency:
        Final drive mean efficiency [-].
    :type final_drive_mean_efficiency: float

    :param gear_box_mean_efficiency_guess:
        Gear box mean efficiency guess [-].
    :type final_drive_mean_efficiency: float

    :param planetary_mean_efficiency:
        Planetary mean efficiency [-].
    :type planetary_mean_efficiency: float

    :param belt_mean_efficiency:
        Belt mean efficiency [-].
    :type belt_mean_efficiency: float

    :param motors_efficiencies:
        Electric motors efficiencies vector.
    :type motors_efficiencies: tuple[float]

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :param hybrid_modes:
        Hybrid mode status (0: EV, 1: Parallel, 2: Serial).
    :type hybrid_modes: numpy.array

    :param motors_maximums_powers:
        Maximum powers of electric motors [kW].
    :type motors_maximums_powers: numpy.array

    :param motors_electric_powers:
        Cumulative motors electric power [kW].
    :type motors_electric_powers: numpy.array

    :param has_motor_p2_planetary:
        Has the vehicle a motor in planetary P2?
    :type has_motor_p2_planetary: bool

    :return:
        Motors electric powers [kW].
    :rtype: tuple[numpy.array]
    """
    drive_line_efficiencies = (
        1.0, final_drive_mean_efficiency, gear_box_mean_efficiency_guess,
        planetary_mean_efficiency, 1.0 if has_motor_p2_planetary else .99,
        belt_mean_efficiency
    )
    return identify_motors_electric_powers(
        HEV(drive_line_efficiencies, motors_efficiencies), hybrid_modes,
        motive_powers, motors_maximums_powers, motors_electric_powers
    )


@sh.add_function(dsp, outputs=['start_stop_activation_time'])
def default_start_stop_activation_time():
    """
    Returns the default start stop activation time threshold [s].

    :return:
        Start-stop activation time threshold [s].
    :rtype: float
    """
    return 0



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and constants to define the start_stop_model selector.
"""
import schedula as sh
from ._core import define_sub_model, _accuracy_score
from ...physical import dsp as _physical

#: Model name.
name = 'control_model'

#: Parameters that constitute the model.
models = [
    'motor_p3_front_efficiency', 'motor_p3_rear_efficiency', 'start_stop_model',
    'motor_p1_maximum_power', 'motor_p1_efficiency', 'gear_box_mean_efficiency',
    'belt_mean_efficiency', 'clutch_tc_mean_efficiency', 'motor_p1_rated_speed',
    'motor_p3_front_rated_speed', 'motor_p2_rated_speed', 'motor_p2_efficiency',
    'start_stop_hybrid_params', 'motor_p0_maximum_power', 'motor_p0_efficiency',
    'motor_p2_planetary_rated_speed', 'motor_p2_planetary_efficiency', 'ecms_s',
    'motor_p3_rear_rated_speed', 'motor_p4_rear_rated_speed', 'planetary_ratio',
    'motor_p2_planetary_maximum_power', 'motor_p4_rear_efficiency', 'is_serial',
    'motor_p3_front_maximum_power', 'motor_p4_front_maximum_power',
    'motor_p3_rear_maximum_power', 'motor_p4_rear_maximum_power',
    'final_drive_mean_efficiency', 'motor_p4_front_rated_speed',
    'planetary_mean_efficiency', 'motor_p4_front_efficiency',
    'motor_p2_maximum_power', 'motor_p0_rated_speed',
]

#: Inputs required to run the model.
inputs = [
    'correct_start_stop_with_gears', 'motor_p0_maximum_power_function', 'gears',
    'motor_p1_maximum_power_function', 'motor_p3_front_maximum_powers', 'times',
    'min_time_engine_on_after_start', 'start_stop_activation_time', 'is_hybrid',
    'motor_p3_rear_maximum_powers', 'motor_p2_maximum_powers', 'has_start_stop',
    'auxiliaries_torque_loss', 'auxiliaries_power_loss', 'motor_p1_speed_ratio',
    'engine_moment_inertia', 'drive_battery_model', 'accelerations', 'fuel_map',
    'motor_p0_speed_ratio', 'gear_box_speeds_in', 'starter_model', 'velocities',
    'motor_p4_rear_maximum_powers', 'is_cycle_hot', 'dcdc_converter_efficiency',
    'after_treatment_cooling_duration', 'idle_engine_speed', 'full_load_curve',
    'after_treatment_warm_up_duration', 'motor_p2_planetary_speed_ratio',
    'motor_p2_planetary_maximum_power_function', 'final_drive_speeds_in',
    'motor_p4_front_maximum_powers', 'motive_powers',
]

#: Relevant outputs of the model.
outputs = ['on_engine', 'engine_starts']

#: Targets to compare the outputs of the model.
targets = outputs

#: Weights coefficients to compute the model score.
weights = sh.map_list(targets, -1, -1)

#: Metrics to compare outputs with targets.
metrics = sh.map_list(targets, *([_accuracy_score] * 2))

#: Bottom score limits to raise the warnings.
dn_limit = sh.map_list(targets, 0.7, 0.7)

#: Prediction model.
# noinspection PyProtectedMember
dsp = sh.Blueprint(_physical, inputs, outputs, models)._set_cls(
    define_sub_model
)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Core functions to predict the A/T gear shifting.
"""
import collections
from co2mpas.defaults import dfl


def define_gear_filter(
        change_gear_window_width=dfl.values.change_gear_window_width):
    """
    Defines a gear filter function.

    :param change_gear_window_width:
        Time window used to apply gear change filters [s].
    :type change_gear_window_width: float

    :return:
        Gear filter function.
    :rtype: callable
    """
    import numpy as np
    from co2mpas.utils import median_filter, clear_fluctuations

    def gear_filter(times, gears):
        """
        Filter the gears to remove oscillations.

        :param times:
            Time vector [s].
        :type times: numpy.array

        :param gears:
            Gear vector [-].
        :type gears: numpy.array

        :return:
            Filtered gears [-].
        :rtype: numpy.array
        """

        gears = median_filter(
            times, gears.astype(float), change_gear_window_width
        )
        gears = clear_fluctuations(times, gears, change_gear_window_width)

        return np.asarray(gears, dtype=int)

    return gear_filter


def prediction_gears_gsm(
        correct_gear, gear_filter, gsm, times, velocities, accelerations,
        motive_powers, cycle_type=None, velocity_speed_ratios=None,
        engine_coolant_temperatures=None):
    """
    Predicts gears with a gear shifting model (cmv or gspv or dtgs or mgs) [-].

    :param correct_gear:
        A function to correct the gear predicted.
    :type correct_gear: callable

    :param gear_filter:
        Gear filter function.
    :type gear_filter: callable

    :param cycle_type:
        Cycle type (WLTP or NEDC).
    :type cycle_type: str

    :param velocity_speed_ratios:
        Constant velocity speed ratios of the gear box [km/(h*RPM)].
    :type velocity_speed_ratios: dict[int | float]

    :param gsm:
        A gear shifting model (cmv or gspv or dtgs).
    :type gsm: GSPV | CMV | DTGS

    :param velocities:
        Vehicle velocity [km/h].
    :type velocities: numpy.array

    :param accelerations:
        Vehicle acceleration [m/s2].
    :type accelerations: numpy.array

    :param times:
        Time vector [s].
    :type times: numpy.array, optional

    :param motive_powers:
        Motive power [kW].
    :type motive_powers: numpy.array

    :param engine_coolant_temperatures:
        Engine coolant temperature vector [°C].
    :type engine_coolant_temperatures: numpy.array

    :return:
        Predicted gears.
    :rtype: numpy.array
    """

    if velocity_speed_ratios is not None and cycle_type is not None:
        from . import _upgrade_gsm
        gsm = _upgrade_gsm(gsm, velocity_speed_ratios, cycle_type)

    # noinspection PyArgumentList
    gears = gsm.predict(
        times, velocities, accelerations, motive_powers,
        engine_coolant_temperatures,
        correct_gear=correct_gear, gear_filter=gear_filter
    )
    return gears


# noinspection PyMissingOrEmptyDocstring
class GSMColdHot(collections.OrderedDict):
    def __init__(self, *args, time_cold_hot_transition=0.0):
        super(GSMColdHot, self).__init__(*args)
        self.time_cold_hot_transition = time_cold_hot_transition

    def __repr__(self):
        name = self.__class__.__name__
        items = [(k, v) for k, v in self.items()]
        s = '{}({}, time_cold_hot_transition={})'.format(
            name, items, self.time_cold_hot_transition
        )
        return s.replace('inf', "float('inf')")

    def fit(self, model_class, times, *args):
        import numpy as np
        self.clear()

        b = times <= self.time_cold_hot_transition

        for i in ['cold', 'hot']:
            if b.sum() > 2:
                a = (v[b] if isinstance(v, np.ndarray) else v for v in args)
                self[i] = model_class().fit(*a)
            b = ~b

        if len(self) == 2 and set(self['cold']) == set(self['hot']):
            return self

    # noinspection PyTypeChecker,PyCallByClass
    def predict(self, *args, **kwargs):
        from .cmv import CMV
        return CMV.predict(self, *args, **kwargs)

    def init_gear(self, gears, times, velocities, accelerations, motive_powers,
                  engine_coolant_temperatures=None,
                  correct_gear=lambda g, *args: g):
        from co2mpas.utils import List
        if gears is None:
            gears = List(dtype=int)

        gen = {k: v.init_gear(
            gears, times, velocities, accelerations, motive_powers,
            engine_coolant_temperatures, correct_gear=correct_gear
        ) for k, v in self.items()}

        def _next(i):
            if times[i] < self.time_cold_hot_transition:
                return gen['cold'](i)
            return gen['hot'](i)

        return _next

    def init_speed(self, *args, **kwargs):
        return self['hot'].init_speed(*args, **kwargs)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to compare/select the calibrated models.

Sub-Modules:

.. currentmodule:: co2mpas.core.model.selector

.. autosummary::
    :nosignatures:
    :toctree: selector/

    models
"""
import pkgutil
import functools
import schedula as sh
import os.path as osp
from .models import mdl_selector, calibration_cycles, prediction_cycles

dsp = sh.BlueDispatcher(
    name='Models selector', description='Select the calibrated models.'
)

dsp.add_function(
    function=functools.partial(sh.map_list, calibration_cycles),
    inputs=calibration_cycles,
    outputs=['CO2MPAS_results']
)

MODELS = [
    v.name
    for v in pkgutil.iter_modules([osp.join(osp.dirname(__file__), 'models')])
    if not v.name.startswith('_')
]

dsp.add_data(
    'enable_selector', False, filters=[lambda x: {'enable_selector': x}]
)

for name in MODELS:
    dsp.add_function(
        function=sh.SubDispatch(
            mdl_selector('.models.%s' % name, __name__),
            outputs=['model', 'score'], output_type='list'
        ),
        function_id='%s selector' % name,
        inputs=['CO2MPAS_results', 'enable_selector'],
        outputs=['models', 'scores']
    )


def merge_model_selection(adict):
    """
    Merge models or scores selection.

    :param adict:
        Models or scores selection.
    :type adict: dict

    :return:
        Merged models or scores selection.
    :rtype: dict
    """
    return {' '.join(k.split(' ')[:-1]): v for k, v in adict.items() if v}


dsp.add_data(data_id='models', function=merge_model_selection, wait_inputs=True)
dsp.add_data(data_id='scores', function=merge_model_selection, wait_inputs=True)


# noinspection PyShadowingNames
@sh.add_function(dsp, outputs=['selections'] + list(
    map('models_{}'.format, prediction_cycles)
))
def split_prediction_models(scores, models, default_models):
    """
    Split prediction models.

    :param scores:
        Models score.
    :type scores: dict

    :param models:
        Calibrated models.
    :type models: dict

    :param default_models:
        Default calibrated models.
    :type default_models: dict

    :return:
        Scores and prediction models.
    :rtype: tuple
    """
    sbm, model_sel, par = {}, {}, {}
    for (k, c), v in sh.stack_nested_keys(scores, depth=2):
        r = sh.selector(['models'], v, allow_miss=True)

        for m in r.get('models', ()):
            sh.get_nested_dicts(par, m, 'calibration')[c] = c

        r.update(v.get('score', {}))
        sh.get_nested_dicts(sbm, k)[c] = r
        r = sh.selector(['success'], r, allow_miss=True)
        r = sh.map_dict({'success': 'status'}, r, {'from': c})
        sh.get_nested_dicts(model_sel, k, 'calibration')[c] = r

    p = {i: dict.fromkeys(default_models, 'input') for i in prediction_cycles}

    mdls = {i: default_models.copy() for i in prediction_cycles}

    for k, n in sorted(models.items()):
        d = n.get(sh.NONE, (None, True, {}))

        for i in prediction_cycles:
            c, s, m = n.get(i, d)
            if m:
                s = {'from': c, 'status': s}
                sh.get_nested_dicts(model_sel, k, 'prediction')[i] = s
                mdls[i].update(m)
                p[i].update(dict.fromkeys(m, c))

    for k, v in sh.stack_nested_keys(p, ('prediction',), depth=2):
        sh.get_nested_dicts(par, k[-1], *k[:-2])[k[-2]] = v

    s = {
        'param_selections': par,
        'model_selections': model_sel,
        'score_by_model': sbm,
        'scores': scores
    }
    return (s,) + tuple(mdls.get(k, {}) for k in prediction_cycles)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to write CO2MPAS output data.

Sub-Modules:

.. currentmodule:: co2mpas.core.write

.. autosummary::
    :nosignatures:
    :toctree: write/

    convert
    excel
"""
import os
import logging
import os.path as osp
import schedula as sh
from .excel import write_to_excel
from .convert import convert2df
from ... import default_start_time, default_timestamp
from co2mpas._version import __version__
from co2mpas.utils import check_first_arg, check_first_arg_false

try:
    from co2mpas_dice import dsp as _dice
except ImportError:
    _dice = None

log = logging.getLogger(__name__)
dsp = sh.BlueDispatcher(
    name='write', description='Write the outputs of the CO2MPAS model.'
)

dsp.add_func(default_start_time, outputs=['start_time'])
dsp.add_func(convert2df, outputs=['dfs'])


@sh.add_function(dsp, outputs=['output_template'], weight=sh.inf(1, 0))
def default_output_template():
    """
    Returns the default template output.

    :return:
        Template output.
    :rtype: str
    """
    from pkg_resources import resource_filename
    return resource_filename('co2mpas', 'templates/output_template.xlsx')


dsp.add_func(write_to_excel, outputs=['excel_output'])

dsp.add_function(
    function=sh.add_args(sh.bypass),
    inputs=['type_approval_mode', 'excel_output'],
    outputs=['output_file'],
    input_domain=check_first_arg_false
)

dsp.add_func(default_timestamp, outputs=['timestamp'])


def default_output_file_name(
        output_folder, vehicle_name, timestamp, ext='xlsx'):
    """
    Returns the output file name.

    :param output_folder:
        Output folder.
    :type output_folder: str

    :param vehicle_name:
        Vehicle name.
    :type vehicle_name: str

    :param timestamp:
        Run timestamp.
    :type timestamp: str

    :param ext:
        File extension.
    :type ext: str | None

    :return:
        Output file name.
    :rtype: str

    """
    fp = osp.join(output_folder, '%s-%s' % (timestamp, vehicle_name))
    if ext is not None:
        fp = '%s.%s' % (fp, ext)
    return fp


dsp.add_function(
    function=sh.add_args(default_output_file_name),
    inputs=['type_approval_mode', 'output_folder', 'vehicle_name', 'timestamp'],
    outputs=['output_file_name'],
    input_domain=check_first_arg_false
)

if _dice is not None:
    dsp.add_data('co2mpas_version', __version__)
    _out, _inp = ['output_file_name', 'output_file'], [
        'excel_input', 'base', 'start_time', 'excel_output', 'output_folder',
        'report', 'encryption_keys', 'meta', 'sign_key', 'dice', 'timestamp',
        'co2mpas_version', 'flag'
    ]

    # noinspection PyProtectedMember,PyTypeChecker
    dsp.add_function(
        function=sh.Blueprint(
            sh.SubDispatchFunction(_dice, inputs=_inp, outputs=_out)
        )._set_cls(sh.add_args),
        function_id='write_ta_output',
        description='Write ta output file.',
        inputs=['type_approval_mode', 'input_file'] + _inp[1:],
        outputs=_out,
        input_domain=check_first_arg
    )
else:
    dsp.add_data('base', description='Base data.')
    dsp.add_data('excel_input', description='Excel input file.')
    dsp.add_data('encryption_keys', description='Encryption keys for TA mode.')
    dsp.add_data('meta', description='Meta data.')
    dsp.add_data('sign_key', description='User signature key for TA mode.')


@sh.add_function(dsp)
def save_output_file(output_file, output_file_name):
    """
    Save output file.

    :param output_file_name:
        Output file name.
    :type output_file_name: str

    :param output_file:
        Output file.
    :type output_file: io.BytesIO
    """
    output_file.seek(0)
    os.makedirs(osp.dirname(output_file_name), exist_ok=True)
    with open(output_file_name, 'wb') as f:
        f.write(output_file.read())
    log.info('CO2MPAS output written into (%s).', output_file_name)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Functions and `dsp` model to model the vehicle control strategy.

Sub-Modules:

.. currentmodule:: co2mpas.core.model.physical.control

.. autosummary::
    :nosignatures:
    :toctree: control/

    conventional
    hybrid
"""
import numpy as np
import schedula as sh
from co2mpas.defaults import dfl
from .conventional import dsp as _conventional
from .hybrid import dsp as _hybrid

dsp = sh.BlueDispatcher(
    name='Control', description='Models the vehicle control strategy.'
)

dsp.add_data(
    'min_time_engine_on_after_start', dfl.values.min_time_engine_on_after_start
)


@sh.add_function(dsp, outputs=['on_engine'])
def identify_on_engine(
        times, engine_speeds_out, idle_engine_speed,
        min_time_engine_on_after_start):
    """
    Identifies if the engine is on [-].

    :param times:
        Time vector [s].
    :type times: numpy.array

    :param engine_speeds_out:
        Engine speed [RPM].
    :type engine_speeds_out: numpy.array

    :param idle_engine_speed:
        Idle engine speed and its standard deviation [RPM].
    :type idle_engine_speed: (float, float)

    :param min_time_engine_on_after_start:
        Minimum time of engine on after a start [s].
    :type min_time_engine_on_after_start: float

    :return:
        If the engine is on [-].
    :rtype: numpy.array
    """

    on_engine = engine_speeds_out > idle_engine_speed[0] - idle_engine_speed[1]
    mask = np.where(identify_engine_starts(on_engine))[0] + 1
    ts = np.asarray(times[mask], dtype=float)
    ts += min_time_engine_on_after_start + dfl.EPS
    for i, j in np.column_stack((mask, np.searchsorted(times, ts))):
        on_engine[i:j] = True

    return on_engine


@sh.add_function(dsp, outputs=['engine_starts'])
def identify_engine_starts(on_engine):
    """
    Identifies when the engine starts [-].

    :param on_engine:
        If the engine is on [-].
    :type on_engine: numpy.array

    :return:
        When the engine starts [-].
    :rtype: numpy.array
    """
    engine_starts = np.zeros_like(on_engine, dtype=bool)
    engine_starts[:-1] = on_engine[1:] & (on_engine[:-1] != on_engine[1:])
    return engine_starts


# noinspection PyMissingOrEmptyDocstring
def is_hybrid(kwargs):
    return kwargs.get('is_hybrid')


# noinspection PyMissingOrEmptyDocstring
def is_not_hybrid(kwargs):
    return not kwargs.get('is_hybrid', True)


dsp.add_dispatcher(
    include_defaults=True,
    dsp=_conventional,
    dsp_id='conventional_control',
    inputs=(
        'gear_box_type', 'start_stop_activation_time', 'accelerations', 'times',
        'idle_engine_speed', 'min_time_engine_on_after_start', 'engine_starts',
        'has_start_stop', 'correct_start_stop_with_gears', 'start_stop_model',
        'gear_box_speeds_in', 'velocities', 'on_engine', 'gears',
        {'is_hybrid': sh.SINK}
    ),
    outputs=(
        'correct_start_stop_with_gears', 'engine_speeds_out_hot', 'on_engine',
        'start_stop_activation_time', 'start_stop_model', 'hybrid_modes',
    ),
    input_domain=is_not_hybrid
)

dsp.add_dispatcher(
    include_defaults=True,
    dsp=_hybrid,
    dsp_id='hybrid_control',
    inputs=(
        'clutch_tc_mean_efficiency', 'idle_engine_speed', 'motor_p0_efficiency',
        'motor_p0_maximum_power', 'drive_battery_model', 'belt_mean_efficiency',
        'min_time_engine_on_after_start', 'start_stop_activation_time', 'times',
        'drive_battery_state_of_charges', 'motor_p2_maximum_powers', 'fuel_map',
        'engine_thermostat_temperature', 'motor_p2_efficiency', 'accelerations',
        'motor_p0_maximum_power_function', 'on_engine', 'engine_moment_inertia',
        'motor_p1_maximum_power_function', 'ecms_s', 'start_stop_hybrid_params',
        'motor_p1_speed_ratio', 'auxiliaries_power_loss', 'motor_p1_efficiency',
        'final_drive_mean_efficiency', 'motors_electric_powers', 'is_cycle_hot',
        'auxiliaries_torque_loss', 'dcdc_converter_efficiency', 'motive_powers',
        'motor_p2_electric_powers', 'motor_p0_speed_ratio', 'engine_speeds_out',
        'motor_p3_rear_efficiency', 'motor_p1_maximum_power', 'full_load_curve',
        'motor_p3_rear_maximum_powers', 'motor_p0_electric_powers', 'is_serial',
        'after_treatment_warm_up_duration', 'engine_powers_out', 'hybrid_modes',
        'motor_p2_maximum_power', 'engine_speeds_out_hot', 'gear_box_speeds_in',
        'gear_box_mean_efficiency', 'motor_p1_electric_powers', 'starter_model',
        'motor_p2_planetary_maximum_power_function', 'motor_p4_rear_efficiency',
        'motor_p3_front_efficiency', 'final_drive_speeds_in', 'planetary_ratio',
        'after_treatment_cooling_duration', 'motor_p2_planetary_maximum_power',
        'motor_p2_planetary_electric_powers', 'after_treatment_warm_up_phases',
        'gear_box_mean_efficiency_guess', 'motor_p3_front_electric_powers',
        'motor_p2_planetary_speed_ratio', 'motor_p4_front_electric_powers',
        'motor_p2_planetary_efficiency', 'motor_p4_rear_electric_powers',
        'motor_p3_front_maximum_powers', 'motor_p3_rear_electric_powers',
        'motor_p4_front_maximum_powers', 'motor_p4_rear_maximum_powers',
        'motor_p4_front_maximum_power', 'motor_p3_front_maximum_power',
        'engine_temperatures', 'motor_p3_rear_maximum_power',
        'motor_p4_rear_maximum_power', 'motor_p4_front_efficiency',
        'planetary_mean_efficiency', 'has_motor_p2_planetary',
        'min_engine_on_speed', {'is_hybrid': sh.SINK},
    ),
    outputs=(
        'motor_p3_front_electric_powers', 'motor_p2_electric_powers', 'ecms_s',
        'motor_p4_front_electric_powers', 'engine_speeds_base', 'hybrid_modes',
        'after_treatment_warm_up_phases', 'engine_speeds_out_hot', 'on_engine',
        'motor_p2_planetary_electric_powers', 'motor_p3_rear_electric_powers',
        'motor_p4_rear_electric_powers', 'start_stop_hybrid_params',
        'motor_p1_electric_powers', 'motor_p0_electric_powers',
        'force_on_engine', 'start_stop_activation_time'
    ),
    input_domain=is_hybrid
)



# -*- coding: utf-8 -*-
#
# Copyright 2015-2022 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
"""
Defines the file processing chain model `dsp`.

.. currentmodule:: co2mpas

.. autosummary::
    :nosignatures:
    :toctree: toctree/co2mpas/

    ~core
    ~cli
    ~utils
    ~defaults
"""
import os
import tqdm
import logging
import webbrowser
import os.path as osp
import schedula as sh
from co2mpas._version import *

log = logging.getLogger(__name__)
dsp = sh.BlueDispatcher(name='process')


def init_conf(inputs):
    """
    Initialize CO2MPAS model configurations.

    :param inputs:
         Initialization inputs.
    :type inputs: dict | schedula.Token

    :return:
        Initialization inputs.
    :rtype: dict | schedula.Token
    """
    if inputs is not sh.NONE and inputs.get('model_conf'):
        from co2mpas.defaults import dfl
        dfl.load(inputs['model_conf'])
        log.info('Model configuration file (%s) loaded.' % inputs['model_conf'])
    return inputs


dsp.add_data(sh.START, filters=[init_conf, lambda x: sh.NONE])


@sh.add_function(dsp, outputs=['demo'])
def save_demo_files(output_folder):
    """
    Save CO2MPAS demo files.

    :param output_folder:
        Output folder.
    :type output_folder: str
    """
    import glob
    from shutil import copy2
    from pkg_resources import resource_filename
    os.makedirs(output_folder or '.', exist_ok=True)
    for src in glob.glob(resource_filename('co2mpas', 'demos/*.xlsx')):
        copy2(src, osp.join(output_folder, osp.basename(src)))
    log.info('CO2MPAS demos written into (%s).', output_folder)


@sh.add_function(dsp, outputs=['template'])
def save_co2mpas_template(output_file, template_type):
    """
    Save CO2MPAS input template.

    :param output_file:
        Output file.
    :type output_file: str

    :param template_type:
        Template type.
    :type template_type: str
    """
    from shutil import copy2
    from pkg_resources import resource_filename
    src = resource_filename(
        'co2mpas', 'templates/%s_template.xlsx' % template_type
    )
    os.makedirs(osp.dirname(output_file) or '.', exist_ok=True)
    copy2(src, output_file)
    log.info('CO2MPAS input template written into (%s).', output_file)


@sh.add_function(dsp, outputs=['conf'])
def save_co2mpas_conf(output_file):
    """
    Save CO2MPAS model configurations.

    :param output_file:
        Output file.
    :type output_file: str
    """
    from co2mpas.defaults import dfl
    os.makedirs(osp.dirname(output_file) or '.', exist_ok=True)
    dfl.dump(output_file)
    log.info('CO2MPAS model configurations written into (%s).', output_file)


@sh.add_function(dsp, outputs=['core_model'])
def register_core():
    """
    Register core model.

    :return:
        CO2MPAS core model.
    :rtype: schedula.Dispatcher
    """
    from .core import dsp
    return dsp.register(memo={})


@sh.add_function(dsp, outputs=['sitemap'])
def plot_core(plot_model, core_model):
    """
    Plot CO2MPAS core model.

    :param plot_model:
        Open model-plot in browser, after run finished.
    :type plot_model: bool

    :param core_model:
        CO2MPAS core model.
    :type core_model: schedula.Dispatcher

    :return:
        Sitemap to plot.
    :rtype: schedula.utils.drw.SiteMap
    """
    if plot_model:
        return core_model.plot(view=False)
    return sh.NONE


@sh.add_function(dsp, outputs=['sitemap'])
def plot_solutions(plot_workflow, solutions):
    """
    Plot CO2MPAS solutions.

    :param plot_workflow:
        Open workflow-plot in browser, after run finished.
    :type plot_workflow: bool

    :param solutions:
        All model solutions.
    :type solutions: list[schedula.Solution]

    :return:
        Sitemap to plot.
    :rtype: schedula.utils.drw.SiteMap
    """
    if plot_workflow:
        from schedula.utils.drw import SiteMap
        sitemap = SiteMap()
        for i, s in enumerate(solutions):
            sitemap.add_items(s, workflow=True)._name = s.get(
                'vehicle_name', 'Solution %d' % i
            )
        return sitemap
    return sh.NONE


@sh.add_function(dsp, outputs=['site'])
def run_sitemap(sitemap, cache_folder, host, port):
    """
    Run sitemap.

    :param sitemap:
        Sitemap to plot.
    :type sitemap: schedula.utils.drw.SiteMap

    :param cache_folder:
        Folder to save temporary html files.
    :type cache_folder: str

    :param host:
        Hostname to listen on.
    :type host: str

    :param port:
        Port of the webserver.
    :type port: int
    :return:
    """
    import flask
    import datetime
    import numpy as np
    flask.Flask.send_file_max_age_default = datetime.timedelta(seconds=0)
    np.set_printoptions(threshold=np.inf)
    site = sitemap.site(cache_folder, host=host, port=port).run()
    webbrowser.open(site.url)
    return site


@sh.add_function(dsp, outputs=['plot'], weight=sh.inf(101, 0))
def wait_site(site):
    """
    Pause for site shutdown.
    """

    import time
    try:
        while True:
            time.sleep(1)
    except (KeyboardInterrupt, SystemExit):
        pass
    site.shutdown()


def _yield_files(*paths, cache=None, ext=(
        'co2mpas.ta', 'dice.ta', 'jet.ta', 'co2mpas', 'xlsx', 'dill', 'xls'
)):
    import glob
    cache = set() if cache is None else cache
    for path in paths:
        path = osp.abspath(path)
        if path in cache or osp.basename(path).startswith('~'):
            continue
        cache.add(path)
        if osp.isdir(path):
            yield from _yield_files(
                *filter(osp.isfile, glob.glob(osp.join(path, '*'))), cache=cache
            )
        elif osp.isfile(path) and path.lower().endswith(ext):
            yield path
        else:
            log.info('Skipping file "%s".' % path)


class _ProgressBar(tqdm.tqdm):
    def __init__(self, *args, _format_meter=None, **kwargs):
        if _format_meter:
            self._format_meter = _format_meter
        super(_ProgressBar, self).__init__(*args, **kwargs)

    @staticmethod
    def _format_meter(bar, data):
        return '%s: Processing %s\n' % (bar, data)

    # noinspection PyMissingOrEmptyDocstring
    def format_meter(self, n, *args, **kwargs):
        bar = super(_ProgressBar, self).format_meter(n, *args, **kwargs)
        try:
            return self._format_meter(bar, self.iterable[n])
        except IndexError:
            return bar


@sh.add_function(dsp, outputs=['start_time'])
def default_start_time():
    """
    Returns the default run start time.

    :return:
        Run start time.
    :rtype: datetime.datetime
    """
    import datetime
    return datetime.datetime.today()


@sh.add_function(dsp, outputs=['timestamp'])
def default_timestamp(start_time):
    """
    Returns the default timestamp.

    :param start_time:
        Run start time.
    :type start_time: datetime.datetime

    :return:
        Run timestamp.
    :rtype: str
    """
    return start_time.strftime('%Y%m%d_%H%M%S')


@sh.add_function(dsp, outputs=['core_solutions'])
def run_core(core_model, cmd_flags, timestamp, input_files, **kwargs):
    """
    Run core model.

    :param core_model:
        CO2MPAS core model.
    :type core_model: schedula.Dispatcher

    :param cmd_flags:
        Command line options.
    :type cmd_flags: dict

    :param timestamp:
        Run timestamp.
    :type timestamp: str

    :param input_files:
        List of input files and/or folders.
    :type input_files: iterable

    :return:
        Core model solutions.
    :rtype: dict[str, schedula.Solution]
    """
    solutions, it = {}, list(_yield_files(*input_files))
    if it:
        for fp in _ProgressBar(it):
            solutions[fp] = core_model(dict(
                input_file_name=fp, cmd_flags=cmd_flags, timestamp=timestamp
            ), **kwargs)
    return solutions


def _define_inputs(sol, inputs):
    kw = dict(
        sources=inputs, check_inputs=False, graph=sol.dsp.dmap,
        _update_links=False
    )
    keys = set(sol) - set(sol.dsp.get_sub_dsp_from_workflow(**kw).data_nodes)
    return sh.combine_dicts({k: sol[k] for k in keys}, inputs)


def _format_meter(bar, row):
    return '%s: Processing %s (%s)\n' % (bar, row['id'], row['base'])


def _run_variations(plan, bases, core_model, timestamp):
    for r in _ProgressBar(plan, _format_meter=_format_meter):
        sol, data = bases[r['base']], r['data']
        if 'solution' in sol:
            s = sol['solution']
            base = _define_inputs(s, sh.combine_nested_dicts(sh.selector(
                data, s, allow_miss=True
            ), data))
        elif 'base' in sol:
            base = sh.combine_nested_dicts(sol['base'], data, depth=2)
        else:
            continue

        for i, d in base.items():
            if hasattr(d, 'items'):
                base[i] = {k: v for k, v in d.items() if v is not sh.EMPTY}

        sol = core_model(_define_inputs(sol, dict(
            base=base,
            vehicle_name='-'.join((str(r['id']), sol['vehicle_name'])),
            timestamp=timestamp
        )))

        summary, keys = {}, {
            tuple(k.split('.')[:0:-1]) for k in base if k.startswith('output.')
        }
        for k, v in data.items():
            k = ('plan %s' % k).split('.')[::-1]
            sh.get_nested_dicts(summary, *k).update(v)

        for k, v in sh.stack_nested_keys(sol['summary'], depth=3):
            if k[:-1] not in keys:
                sh.get_nested_dicts(summary, *k).update(v)
        sol['summary'] = summary
        yield sol


@sh.add_function(dsp, outputs=['solutions'])
def run_plan(core_solutions, core_model, cmd_flags, timestamp):
    """
    Run simulation plans.

    :param core_solutions:
        Core model solutions.
    :type core_solutions:  dict[str, schedula.Solution]

    :param core_model:
        CO2MPAS core model.
    :type core_model: schedula.Dispatcher

    :param cmd_flags:
        Command line options.
    :type cmd_flags: dict

    :param timestamp:
        Run timestamp.
    :type timestamp: str

    :return:
        All model solutions.
    :rtype: list[schedula.Solution]
    """
    bases = core_solutions.copy()
    plan = sum((sol.get('plan', []) for sol in bases.values()), [])
    # Run base.
    fp = {r['base'] for r in plan if r['run_base']} - set(bases)
    bases.update(run_core(core_model, cmd_flags, timestamp, fp))
    solutions = list(bases.values())
    # Load inputs.
    fp = {r['base'] for r in plan if not r['run_base']} - set(bases) - fp
    bases.update(run_core(
        core_model, cmd_flags, timestamp, fp, outputs=['base', 'vehicle_name']
    ))
    if plan:
        solutions.extend(_run_variations(plan, bases, core_model, timestamp))
    return solutions


@sh.add_function(dsp, outputs=['summary'])
def get_summary(solutions):
    """
    Extract summary data from model solutions.

    :param solutions:
        All model solutions.
    :type solutions: list[schedula.Solution]

    :return:
        Summary data.
    :rtype: list
    """
    return [sh.combine_dicts(
        dict(sh.stack_nested_keys(sol.get('summary', {}), depth=4)), base={
            'id': sol['vehicle_name'],
            'base': sol['input_file_name']
        }
    ) for sol in solutions]


@sh.add_function(dsp, outputs=['output_summary_file'])
def define_output_summary_file(cmd_flags, timestamp):
    """
    Defines the output summary file path.

    :param cmd_flags:
        Command line options.
    :type cmd_flags: dict

    :param timestamp:
        Run timestamp.
    :type timestamp: str

    :return:
        Output summary file path.
    :rtype: str
    """
    fp = cmd_flags.get('output_folder', './outputs')
    return osp.join(fp, '%s-summary.xlsx' % timestamp)


@sh.add_function(dsp, outputs=['run'], input_domain=lambda *args: bool(args[0]))
def save_summary(summary, output_summary_file, start_time):
    """
    Save CO2MPAS model configurations.

    :param summary:
        Summary data.
    :type summary: list

    :param output_summary_file:
        Output summary file path.
    :type output_summary_file: str

    :param start_time:
        Run start time.
    :type start_time: datetime.datetime
    """
    import pandas as pd
    # noinspection PyProtectedMember
    from .core.write.convert import _co2mpas_info2df, _add_units, _sort_key
    df = pd.DataFrame(summary)
    df.set_index(['id', 'base'], inplace=True)
    df = df.reindex(columns=sorted(
        df.columns,
        key=lambda x: _sort_key(x, p_keys=('cycle', 'stage', 'usage', 'param'))
    ))
    if not df.columns.empty:
        df.columns = pd.MultiIndex.from_tuples(_add_units(
            df.columns, short=False
        ))

    os.makedirs(osp.dirname(output_summary_file) or '.', exist_ok=True)
    with pd.ExcelWriter(output_summary_file) as writer:
        df.to_excel(writer, 'summary')
        _co2mpas_info2df(start_time).to_excel(writer, 'proc_info')
    log.info('CO2MPAS summary written into (%s)...', output_summary_file)


@sh.add_function(dsp, outputs=['done'], weight=sh.inf(100, 0))
def log_done(start_time):
    """
    Logs the overall execution time.

    :param start_time:
        Run start time.
    :type start_time: datetime.datetime

    :return:
        Execution time [s].
    :rtype:
    """
    import datetime
    sec = (datetime.datetime.today() - start_time).total_seconds()
    log.info('Done! [%.2f sec]' % sec)
    return sec
